# -*- coding: utf-8 -*-


from jax import lax
from jax.tree_util import tree_flatten, tree_unflatten
try:
  from jax.errors import UnexpectedTracerError
except ImportError:
  from jax.core import UnexpectedTracerError

from brainpy import errors
from brainpy.math.jaxarray import JaxArray
from brainpy.math.numpy_ops import as_device_array

__all__ = [
  'make_loop',
  'make_while',
  'make_cond',
]


def _get_scan_info(f, dyn_vars, out_vars=None, has_return=False):
  # iterable variables
  if isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(
      f'"dyn_vars" does not support {type(dyn_vars)}, '
      f'only support dict/list/tuple of {JaxArray.__name__}')
  for v in dyn_vars:
    if not isinstance(v, JaxArray):
      raise ValueError(
        f'brainpy.math.jax.make_loop only support '
        f'{JaxArray.__name__}, but got {type(v)}')

  # outputs
  if out_vars is None:
    out_vars = ()
    _, tree = tree_flatten(out_vars)
  elif isinstance(out_vars, JaxArray):
    _, tree = tree_flatten(out_vars)
    out_vars = (out_vars,)
  elif isinstance(out_vars, dict):
    _, tree = tree_flatten(out_vars)
    out_vars = tuple(out_vars.values())
  elif isinstance(out_vars, (tuple, list)):
    _, tree = tree_flatten(out_vars)
    out_vars = tuple(out_vars)
  else:
    raise ValueError(
      f'"out_vars" does not support {type(out_vars)}, '
      f'only support dict/list/tuple of {JaxArray.__name__}')

  # functions
  if has_return:
    def fun2scan(dyn_values, x):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      results = f(x)
      dyn_values = [v.value for v in dyn_vars]
      out_values = [v.value for v in out_vars]
      return dyn_values, (out_values, results)

  else:
    def fun2scan(dyn_values, x):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      f(x)
      dyn_values = [v.value for v in dyn_vars]
      out_values = [v.value for v in out_vars]
      return dyn_values, out_values

  return fun2scan, dyn_vars, tree


def make_loop(body_fun, dyn_vars, out_vars=None, has_return=False):
  """Make a for-loop function, which iterate over inputs.

  Examples
  --------

  >>> import brainpy.math as bm
  >>>
  >>> a = bm.zeros(1)
  >>> def f(x): a.value += 1.
  >>> loop = bm.make_loop(f, dyn_vars=[a], out_vars=a)
  >>> loop(length=10)
  JaxArray(DeviceArray([[ 1.],
                        [ 2.],
                        [ 3.],
                        [ 4.],
                        [ 5.],
                        [ 6.],
                        [ 7.],
                        [ 8.],
                        [ 9.],
                        [10.]], dtype=float32))
  >>> b = bm.zeros(1)
  >>> def f(x):
  >>>   b.value += 1
  >>>   return b + 1
  >>> loop = bm.make_loop(f, dyn_vars=[b], out_vars=b, has_return=True)
  >>> hist_b, hist_b_plus = loop(length=10)
  >>> hist_b
  JaxArray(DeviceArray([[ 1.],
                        [ 2.],
                        [ 3.],
                        [ 4.],
                        [ 5.],
                        [ 6.],
                        [ 7.],
                        [ 8.],
                        [ 9.],
                        [10.]], dtype=float32))
  >>> hist_b_plus
  JaxArray(DeviceArray([[ 2.],
                        [ 3.],
                        [ 4.],
                        [ 5.],
                        [ 6.],
                        [ 7.],
                        [ 8.],
                        [ 9.],
                        [10.],
                        [11.]], dtype=float32))

  Parameters
  ----------
  body_fun : callable, function
    A function receive one argument. This argument refers to the iterable input ``x``.
  dyn_vars : dict of JaxArray, sequence of JaxArray
    The dynamically changed variables, while iterate between trials.
  out_vars : optional, JaxArray, dict of JaxArray, sequence of JaxArray
    The variables to output their values.
  has_return : bool
    The function has the return values.

  Returns
  -------
  loop_func : callable, function
    The function for loop iteration. This function receives one argument ``xs``, denoting
    the input tensor which interate over the time (note ``body_fun`` receive ``x``).
  """

  fun2scan, dyn_vars, tree = _get_scan_info(f=body_fun,
                                            dyn_vars=dyn_vars,
                                            out_vars=out_vars,
                                            has_return=has_return)

  # functions
  if has_return:
    def call(xs=None, length=None):
      init_values = [v.value for v in dyn_vars]
      try:
        dyn_values, (out_values, results) = lax.scan(
          f=fun2scan, init=init_values, xs=xs, length=length)
      except UnexpectedTracerError as e:
        for v, d in zip(dyn_vars, init_values): v.value = d
        raise errors.JaxTracerError(variables=dyn_vars) from e
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      return tree_unflatten(tree, out_values), results

  else:
    def call(xs):
      init_values = [v.value for v in dyn_vars]
      try:
        dyn_values, out_values = lax.scan(f=fun2scan, init=init_values, xs=xs)
      except UnexpectedTracerError as e:
        for v, d in zip(dyn_vars, init_values): v.value = d
        raise errors.JaxTracerError(variables=dyn_vars) from e
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      return tree_unflatten(tree, out_values)

  return call


def make_while(cond_fun, body_fun, dyn_vars):
  """Make a while-loop function.

  This function is similar to the ``jax.lax.while_loop``. The difference is that,
  if you are using ``JaxArray`` in your while loop codes, this function will help
  you make a easy while loop function. Note: ``cond_fun`` and ``body_fun`` do no
  receive any arguments. ``cond_fun`` shoule return a boolean value. ``body_fun``
  does not support return values.

  Examples
  --------
  >>> import brainpy.math as bm
  >>>
  >>> a = bm.zeros(1)
  >>>
  >>> def cond_f(x): return a[0] < 10
  >>> def body_f(x): a.value += 1.
  >>>
  >>> loop = bm.make_while(cond_f, body_f, dyn_vars=[a])
  >>> loop()
  >>> a
  JaxArray(DeviceArray([10.], dtype=float32))

  Parameters
  ----------
  cond_fun : function, callable
    A function receives one argument, but return a boolean value.
  body_fun : function, callable
    A function receives one argument, without any returns.
  dyn_vars : dict of JaxArray, sequence of JaxArray
    The dynamically changed variables, while iterate between trials.

  Returns
  -------
  loop_func : callable, function
      The function for loop iteration, which receive one argument ``x`` for external input.
  """
  # iterable variables
  if isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(
      f'"dyn_vars" does not support {type(dyn_vars)}, '
      f'only support dict/list/tuple of {JaxArray.__name__}')
  for v in dyn_vars:
    if not isinstance(v, JaxArray):
      raise ValueError(f'brainpy.math.jax.loops only support {JaxArray.__name__}, but got {type(v)}')

  def _body_fun(op):
    dyn_values, static_values = op
    for v, d in zip(dyn_vars, dyn_values): v.value = d
    body_fun(static_values)
    return [v.value for v in dyn_vars], static_values

  def _cond_fun(op):
    dyn_values, static_values = op
    for v, d in zip(dyn_vars, dyn_values): v.value = d
    return as_device_array(cond_fun(static_values))

  def call(x=None):
    dyn_init = [v.value for v in dyn_vars]
    try:
      dyn_values, _ = lax.while_loop(cond_fun=_cond_fun,
                                     body_fun=_body_fun,
                                     init_val=(dyn_init, x))
    except UnexpectedTracerError as e:
      for v, d in zip(dyn_vars, dyn_init): v.value = d
      raise errors.JaxTracerError(variables=dyn_vars) from e
    for v, d in zip(dyn_vars, dyn_values): v.value = d

  return call


def make_cond(true_fun, false_fun, dyn_vars=None):
  """Make a condition (if-else) function.

  Examples
  --------

  >>> import brainpy.math as bm
  >>> a = bm.zeros(2)
  >>> b = bm.ones(2)
  >>>
  >>> def true_f(x):  a.value += 1
  >>> def false_f(x): b.value -= 1
  >>>
  >>> cond = bm.make_cond(true_f, false_f, dyn_vars=[a, b])
  >>> cond(True)
  >>> a, b
  (JaxArray(DeviceArray([1., 1.], dtype=float32)),
   JaxArray(DeviceArray([1., 1.], dtype=float32)))
  >>> cond(False)
  >>> a, b
  (JaxArray(DeviceArray([1., 1.], dtype=float32)),
   JaxArray(DeviceArray([0., 0.], dtype=float32)))

  Parameters
  ----------
  true_fun : callable, function
    A function receives one argument, without any returns.
  false_fun : callable, function
    A function receives one argument, without any returns.
  dyn_vars : dict of JaxArray, sequence of JaxArray
    The dynamically changed variables.

  Returns
  -------
  cond_func : callable, function
      The condictional function receives two arguments: ``pred`` for true/false judgement
      and ``x`` for external input.
  """
  # iterable variables
  if dyn_vars is None:
    dyn_vars = []
  if isinstance(dyn_vars, JaxArray):
    dyn_vars = (dyn_vars, )
  elif isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(
      f'"dyn_vars" does not support {type(dyn_vars)}, '
      f'only support dict/list/tuple of {JaxArray.__name__}')
  for v in dyn_vars:
    if not isinstance(v, JaxArray):
      raise ValueError(
        f'brainpy.math.jax.loops only support '
        f'{JaxArray.__name__}, but got {type(v)}')

  def _true_fun(op):
    dyn_vals, static_vals = op
    for v, d in zip(dyn_vars, dyn_vals): v.value = d
    res = true_fun(static_vals)
    dyn_vals = [v.value for v in dyn_vars]
    return dyn_vals, res

  def _false_fun(op):
    dyn_vals, static_vals = op
    for v, d in zip(dyn_vars, dyn_vals): v.value = d
    res = false_fun(static_vals)
    dyn_vals = [v.value for v in dyn_vars]
    return dyn_vals, res

  def call(pred, x=None):
    old_values = [v.value for v in dyn_vars]
    try:
      dyn_values, res = lax.cond(pred=pred,
                                 true_fun=_true_fun,
                                 false_fun=_false_fun,
                                 operand=(old_values, x))
    except UnexpectedTracerError as e:
      for v, d in zip(dyn_vars, old_values): v.value = d
      raise errors.JaxTracerError(variables=dyn_vars) from e
    for v, d in zip(dyn_vars, dyn_values): v.value = d
    return res

  return call



# -*- coding: utf-8 -*-

from functools import partial
from typing import Union, Callable, Dict, Sequence

import jax
import numpy as np
from jax import linear_util, dtypes, vmap, numpy as jnp
from jax._src.api import (_vjp, _jvp,
                          _check_callable,
                          _check_output_dtype_jacrev, _check_input_dtype_jacrev,
                          _check_output_dtype_jacfwd, _check_input_dtype_jacfwd, )
from jax.api_util import argnums_partial
from jax.errors import UnexpectedTracerError
from jax.tree_util import tree_flatten, tree_unflatten, tree_map, tree_transpose, tree_structure
from jax.util import safe_map

from brainpy import errors
from brainpy.math.jaxarray import JaxArray

__all__ = [
  'grad',  # gradient of scalar function
  'vector_grad',  # gradient of vector/matrix/...
  'jacobian', 'jacrev', 'jacfwd',  # gradient of jacobian
  'hessian',  # gradient of hessian
]


def _make_cls_call_func(grad_func, grad_tree, grad_vars, dyn_vars,
                        argnums, return_value, has_aux):
  # outputs
  def call_func(*args, **kwargs):
    old_grad_vs = [v.value for v in grad_vars]
    old_dyn_vs = [v.value for v in dyn_vars]
    try:
      grads, (outputs, new_grad_vs, new_dyn_vs) = grad_func(old_grad_vs,
                                                            old_dyn_vs,
                                                            *args,
                                                            **kwargs)
    except UnexpectedTracerError as e:
      for v, d in zip(grad_vars, old_grad_vs): v.value = d
      for v, d in zip(dyn_vars, old_dyn_vs): v.value = d
      raise errors.JaxTracerError(variables=dyn_vars + grad_vars) from e
    for v, d in zip(grad_vars, new_grad_vs): v.value = d
    for v, d in zip(dyn_vars, new_dyn_vs): v.value = d

    # check returned grads
    if len(grad_vars) == 0:
      grads = grads[1] if isinstance(argnums, int) else grads[1:]
    else:
      var_grads = grad_tree.unflatten(grads[0])
      if argnums is None:
        grads = var_grads
      else:
        arg_grads = grads[1] if isinstance(argnums, int) else grads[1:]
        grads = (var_grads, arg_grads)

    # check returned value
    if return_value:
      # check aux
      if has_aux:
        return grads, outputs[0], outputs[1]
      else:
        return grads, outputs
    else:
      # check aux
      if has_aux:
        return grads, outputs[1]
      else:
        return grads

  return call_func


def _check_vars(variables):
  if variables is None:
    vars, tree = tree_flatten(variables, is_leaf=lambda a: isinstance(a, JaxArray))
    return vars, tree
  if isinstance(variables, dict):
    variables = dict(variables)
  elif isinstance(variables, (list, tuple)):
    variables = tuple(variables)
  elif isinstance(variables, JaxArray):
    pass
  else:
    raise ValueError
  vars, tree = tree_flatten(variables, is_leaf=lambda a: isinstance(a, JaxArray))
  for v in vars:
    if not isinstance(v, JaxArray):
      raise ValueError(f'"dyn_vars" and "grad_vars" only supports dict '
                       f'of JaxArray, but got {type(v)}: {v}')
  return vars, tree


def _grad_checking(func: Callable,
                   dyn_vars: Union[Dict, Sequence],
                   grad_vars: Union[Dict, Sequence]):
  # check function
  if not callable(func):
    raise ValueError(f'Must be a callable object. But we got {func}')

  # check "vars", make sure it is an instance of TensorCollector
  dyn_vars, _ = _check_vars(dyn_vars)
  grad_vars, grad_tree = _check_vars(grad_vars)

  # check the duplicate in "dyn_vars" and "grad_vars"
  new_dyn_vars = []
  _dyn_var_ids = set()
  for v in dyn_vars:
    if id(v) not in _dyn_var_ids:
      new_dyn_vars.append(v)
      _dyn_var_ids.add(id(v))
  for v in grad_vars:
    if id(v) not in _dyn_var_ids:
      new_dyn_vars.append(v)
      _dyn_var_ids.add(id(v))
  return new_dyn_vars, grad_vars, grad_tree


def _cls_grad(func, grad_vars, dyn_vars, argnums, has_aux=False,
              holomorphic=False, allow_int=False, reduce_axes=()):
  # parameters
  assert isinstance(dyn_vars, (tuple, list))  # tuple/list of JaxArray
  assert isinstance(grad_vars, (tuple, list))  # tuple/list of JaxArray
  assert isinstance(argnums, (tuple, list))  # tuple/list of int

  # gradient functions
  if has_aux:
    @partial(jax.grad, argnums=argnums, has_aux=True, holomorphic=holomorphic,
             allow_int=allow_int, reduce_axes=reduce_axes)
    def grad_func(grad_values, dyn_values, *args, **kwargs):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      for v, d in zip(grad_vars, grad_values): v.value = d
      # Users should return the auxiliary data like::
      # >>> # 1. example of return one data
      # >>> return scalar_loss, data
      # >>> # 2. example of return multiple data
      # >>> return scalar_loss, (data1, data2, ...)
      outputs = func(*args, **kwargs)
      # outputs: [0] is the value for gradient,
      #          [1] is other values for return
      output = outputs[0].value if isinstance(outputs[0], JaxArray) else outputs[0]
      return output, (outputs, [v.value for v in grad_vars], [v.value for v in dyn_vars])
  else:
    @partial(jax.grad,
             argnums=argnums, has_aux=True, holomorphic=holomorphic,
             allow_int=allow_int, reduce_axes=reduce_axes)
    def grad_func(grad_values, dyn_values, *args, **kwargs):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      for v, d in zip(grad_vars, grad_values): v.value = d
      # Users should return the scalar value like this::
      # >>> return scalar_loss
      output = func(*args, **kwargs)
      output2 = output.value if isinstance(output, JaxArray) else output
      return output2, (output, [v.value for v in grad_vars], [v.value for v in dyn_vars])

  return grad_func


def grad(func, grad_vars=None, dyn_vars=None, argnums=None, holomorphic=False,
         allow_int=False, reduce_axes=(), has_aux=None, return_value=False) -> callable:
  """Automatic gradient computation for functions or class objects.

  This gradient function only support scalar return. It creates a function
  which evaluates the gradient of ``func``.

  It's worthy to note that the returns are different for different argument settings (where ``arg_grads`` refers
  to the gradients of "argnums", and ``var_grads`` refers to the gradients of "grad_vars").

  1. When "grad_vars" is None
    - "has_aux=False" + "return_value=False" => ``arg_grads``.
    - "has_aux=True" + "return_value=False" => ``(arg_grads, aux_data)``.
    - "has_aux=False" + "return_value=True" => ``(arg_grads, loss_value)``.
    - "has_aux=True" + "return_value=True" => ``(arg_grads, loss_value, aux_data)``.
  2. When "grad_vars" is not None and "argnums" is None
    - "has_aux=False" + "return_value=False" => ``var_grads``.
    - "has_aux=True" + "return_value=False" => ``(var_grads, aux_data)``.
    - "has_aux=False" + "return_value=True" => ``(var_grads, loss_value)``.
    - "has_aux=True" + "return_value=True" => ``(var_grads, loss_value, aux_data)``.
  3. When "grad_vars" is not None and "argnums" is not None
    - "has_aux=False" + "return_value=False" => ``(var_grads, arg_grads)``.
    - "has_aux=True" + "return_value=False" => ``((var_grads, arg_grads), aux_data)``.
    - "has_aux=False" + "return_value=True" => ``((var_grads, arg_grads), loss_value)``.
    - "has_aux=True" + "return_value=True" => ``((var_grads, arg_grads), loss_value, aux_data)``.

  Let's see some examples below.

  Before start, let's figure out what should be provided as ``grad_vars``?
  And, what should be labeled in ``argnums``?
  Take the following codes as example:

  >>> import brainpy as bp
  >>> import brainpy.math as bm
  >>>
  >>> class Example(bp.Base):
  >>>   def __init__(self):
  >>>     super(Example, self).__init__()
  >>>     self.x = bm.TrainVar(bm.zeros(1))
  >>>     self.y = bm.random.rand(10)
  >>>   def __call__(self, z, v):
  >>>     t1 = self.x * self.y.sum()
  >>>     t2 = bm.tanh(z * v + t1)
  >>>     return t2.mean()
  >>>
  >>> # This code is equivalent to the following function:
  >>>
  >>> x = bm.TrainVar(bm.zeros(1))
  >>> y = bm.random.rand(10)
  >>> def f(z, v):
  >>>   t1 = x * y.sum()
  >>>   t2 = bm.tanh(z * v + t1)
  >>>   return t2.mean()

  Generally speaking, all gradient variables which not provided in arguments should be
  labeled as ``grad_vars``, while all gradient variables provided in the function arguments
  should be declared in ``argnums``.
  In above codes, we try to take gradients of ``self.x`` and arguments ``z`` and ``v``, we should
  call ``brainpy.math.grad`` as:

  >>> f = Example()
  >>> f_grad = bm.grad(f, grad_vars=f.x, argnums=(0, 1))


  Examples
  --------

  Grad for a pure function:

  >>> import brainpy as bp
  >>> grad_tanh = grad(bp.math.tanh)
  >>> print(grad_tanh(0.2))
  0.961043




  Parameters
  ----------
  func : function, Base
    Function to be differentiated. Its arguments at positions specified by
    ``argnums`` should be arrays, scalars, or standard Python containers.
    Argument arrays in the positions specified by ``argnums`` must be of
    inexact (i.e., floating-point or complex) type. It should return a scalar
    (which includes arrays with shape ``()`` but not arrays with shape ``(1,)`` etc.)
  dyn_vars : optional, JaxArray, sequence of JaxArray, dict
    The dynamically changed variables used in ``func``.
  grad_vars : optional, JaxArray, sequence of JaxArray, dict
    The variables in ``func`` to take their gradients.
  argnums : optional, integer or sequence of integers
    Specifies which positional argument(s) to differentiate with respect to (default 0).
  has_aux: optional, bool
    Indicates whether ``fun`` returns a pair where the
    first element is considered the output of the mathematical function to be
    differentiated and the second element is auxiliary data. Default False.
  return_value : bool
    Whether return the loss value.
  holomorphic: optional, bool
    Indicates whether ``fun`` is promised to be
    holomorphic. If True, inputs and outputs must be complex. Default False.
  allow_int: optional, bool
    Whether to allow differentiating with
    respect to integer valued inputs. The gradient of an integer input will
    have a trivial vector-space dtype (float0). Default False.
  reduce_axes: optional, tuple of int
    tuple of axis names. If an axis is listed here, and
    ``fun`` implicitly broadcasts a value over that axis, the backward pass
    will perform a ``psum`` of the corresponding gradient. Otherwise, the
    gradient will be per-example over named axes. For example, if ``'batch'``
    is a named batch axis, ``grad(f, reduce_axes=('batch',))`` will create a
    function that computes the total gradient while ``grad(f)`` will create
    one that computes the per-example gradient.

  Returns
  -------
  func : function
    A function with the same arguments as ``fun``, that evaluates the gradient
    of ``fun``. If ``argnums`` is an integer then the gradient has the same
    shape and type as the positional argument indicated by that integer. If
    argnums is a tuple of integers, the gradient is a tuple of values with the
    same shapes and types as the corresponding arguments. If ``has_aux`` is True
    then a pair of (gradient, auxiliary_data) is returned.
  """

  dyn_vars, grad_vars, grad_tree = _grad_checking(func, dyn_vars, grad_vars)
  # dyn_vars -> TensorCollector
  # grad_vars -> TensorCollector
  has_aux = False if has_aux is None else has_aux

  # gradient
  if len(dyn_vars) == 0 and len(grad_vars) == 0:
    argnums = 0 if argnums is None else argnums
    if return_value:
      grad_func = jax.value_and_grad(fun=func,
                                     argnums=argnums,
                                     has_aux=has_aux,
                                     holomorphic=holomorphic,
                                     allow_int=allow_int,
                                     reduce_axes=reduce_axes)

      def call_func(*args, **kwargs):
        result = grad_func(*args, **kwargs)
        if has_aux:
          (ans, aux), g = result
          return g, ans, aux
        else:
          ans, g = result
          return g, ans

    else:
      # has_aux = True: g, aux
      # has_aux = False: g
      call_func = jax.grad(fun=func,
                           argnums=argnums,
                           has_aux=has_aux,
                           holomorphic=holomorphic,
                           allow_int=allow_int,
                           reduce_axes=reduce_axes)
  else:
    # argnums
    _argnums, _ = tree_flatten(argnums)
    _argnums = tuple(a + 2 for a in _argnums)
    if argnums is None and len(grad_vars) == 0:
      raise errors.MathError('We detect no require to compute gradients because '
                             '"grad_vars" is None and "argnums" is also None. '
                             'Please provide one of them.')
    # computation
    grad_func = _cls_grad(func=func,
                          grad_vars=grad_vars,
                          dyn_vars=dyn_vars,
                          argnums=(0,) + _argnums,
                          has_aux=has_aux,
                          holomorphic=holomorphic,
                          allow_int=allow_int,
                          reduce_axes=reduce_axes)

    call_func = _make_cls_call_func(grad_func=grad_func,
                                    grad_tree=grad_tree,
                                    grad_vars=grad_vars,
                                    dyn_vars=dyn_vars,
                                    argnums=argnums,
                                    return_value=return_value,
                                    has_aux=has_aux)

  return call_func  # Finally, return the callable function


def _unravel_array_into_pytree(pytree, axis, arr, is_leaf=None):
  leaves, treedef = tree_flatten(pytree, is_leaf=is_leaf)
  axis = axis % arr.ndim
  shapes = [arr.shape[:axis] + np.shape(l) + arr.shape[axis + 1:] for l in leaves]
  parts = arr.split(np.cumsum(safe_map(np.size, leaves[:-1])), axis)
  reshaped_parts = [x.reshape(shape) for x, shape in zip(parts, shapes)]
  return tree_unflatten(treedef, reshaped_parts, )


def _std_basis(pytree):
  leaves, _ = tree_flatten(pytree)
  ndim = sum(safe_map(np.size, leaves))
  dtype = dtypes.result_type(*leaves)
  flat_basis = jax.numpy.eye(ndim, dtype=dtype)
  return _unravel_array_into_pytree(pytree, 1, flat_basis)


_isleaf = lambda x: isinstance(x, JaxArray)


def _jacrev(fun, argnums=0, holomorphic=False, allow_int=False, has_aux=False, return_value=False):
  _check_callable(fun)

  def jacfun(*args, **kwargs):
    f = linear_util.wrap_init(fun, kwargs)
    f_partial, dyn_args = argnums_partial(f, argnums, args, require_static_args_hashable=False)
    tree_map(partial(_check_input_dtype_jacrev, holomorphic, allow_int), dyn_args)
    if has_aux:
      y, pullback, aux = _vjp(f_partial, *dyn_args, has_aux=True)
    else:
      y, pullback = _vjp(f_partial, *dyn_args, has_aux=False)
    tree_map(partial(_check_output_dtype_jacrev, holomorphic), y)
    jac = vmap(pullback)(_std_basis(y))
    jac = jac[0] if isinstance(argnums, int) else jac
    example_args = dyn_args[0] if isinstance(argnums, int) else dyn_args
    jac_tree = tree_map(partial(_unravel_array_into_pytree, y, 0, is_leaf=_isleaf), jac, is_leaf=_isleaf)
    jac = tree_transpose(tree_structure(example_args), tree_flatten(y, is_leaf=_isleaf)[1], jac_tree)
    if return_value:
      return (jac, y, aux) if has_aux else (jac, y)
    else:
      return (jac, aux) if has_aux else jac

  return jacfun


def _cls_jacrev(func, grad_vars, dyn_vars, argnums,
                holomorphic=False, allow_int=False, has_aux=False):
  # parameters
  assert isinstance(dyn_vars, (tuple, list))  # tuple/list of JaxArray
  assert isinstance(grad_vars, (tuple, list))  # tuple/list of JaxArray
  assert isinstance(argnums, (tuple, list))  # tuple/list of int

  # final functions
  if has_aux:
    @partial(_jacrev, argnums=argnums, holomorphic=holomorphic,
             allow_int=allow_int, has_aux=True)
    def grad_func(grad_values, dyn_values, *args, **kwargs):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      for v, d in zip(grad_vars, grad_values): v.value = d
      # outputs: [0] is the value for gradient,
      #          [1] is other values for return
      outputs = func(*args, **kwargs)
      output = outputs[0].value if isinstance(outputs[0], JaxArray) else outputs[0]
      return output, (outputs, [v.value for v in grad_vars], [v.value for v in dyn_vars])

  else:
    @partial(_jacrev, argnums=argnums, holomorphic=holomorphic,
             allow_int=allow_int, has_aux=True)
    def grad_func(grad_values, dyn_values, *args, **kwargs):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      for v, d in zip(grad_vars, grad_values): v.value = d
      outputs = func(*args, **kwargs)
      output = outputs.value if isinstance(outputs, JaxArray) else outputs
      return output, (outputs, [v.value for v in grad_vars], [v.value for v in dyn_vars])

  return grad_func


def jacrev(func, grad_vars=None, dyn_vars=None, argnums=None, holomorphic=False,
           allow_int=False, has_aux=None, return_value=False):
  """Extending automatic Jacobian (reverse-mode) of ``func`` to classes.

  This function extends the JAX official ``jacrev`` to make automatic jacobian
  computation on functions and class functions. Moreover, it supports returning
  value ("return_value") and returning auxiliary data ("has_aux").

  Same as `brainpy.math.grad <./brainpy.math.autograd.grad.html>`_, the returns are
  different for different argument settings in ``brainpy.math.jacrev``.

  1. When "grad_vars" is None
    - "has_aux=False" + "return_value=False" => ``arg_grads``.
    - "has_aux=True" + "return_value=False" => ``(arg_grads, aux_data)``.
    - "has_aux=False" + "return_value=True" => ``(arg_grads, loss_value)``.
    - "has_aux=True" + "return_value=True" => ``(arg_grads, loss_value, aux_data)``.
  2. When "grad_vars" is not None and "argnums" is None
    - "has_aux=False" + "return_value=False" => ``var_grads``.
    - "has_aux=True" + "return_value=False" => ``(var_grads, aux_data)``.
    - "has_aux=False" + "return_value=True" => ``(var_grads, loss_value)``.
    - "has_aux=True" + "return_value=True" => ``(var_grads, loss_value, aux_data)``.
  3. When "grad_vars" is not None and "argnums" is not None
    - "has_aux=False" + "return_value=False" => ``(var_grads, arg_grads)``.
    - "has_aux=True" + "return_value=False" => ``((var_grads, arg_grads), aux_data)``.
    - "has_aux=False" + "return_value=True" => ``((var_grads, arg_grads), loss_value)``.
    - "has_aux=True" + "return_value=True" => ``((var_grads, arg_grads), loss_value, aux_data)``.


  Parameters
  ----------
  func: Function whose Jacobian is to be computed.
  dyn_vars : optional, JaxArray, sequence of JaxArray, dict
    The dynamically changed variables used in ``func``.
  grad_vars : optional, JaxArray, sequence of JaxArray, dict
    The variables in ``func`` to take their gradients.
  has_aux: optional, bool
    Indicates whether ``fun`` returns a pair where the
    first element is considered the output of the mathematical function to be
    differentiated and the second element is auxiliary data. Default False.
  return_value : bool
    Whether return the loss value.
  argnums: Optional, integer or sequence of integers. Specifies which
    positional argument(s) to differentiate with respect to (default ``0``).
  holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
    holomorphic. Default False.
  allow_int: Optional, bool. Whether to allow differentiating with
    respect to integer valued inputs. The gradient of an integer input will
    have a trivial vector-space dtype (float0). Default False.

  """
  dyn_vars, grad_vars, grad_tree = _grad_checking(func, dyn_vars, grad_vars)
  has_aux = False if has_aux is None else has_aux

  if (len(dyn_vars) == 0) and (len(grad_vars) == 0):
    argnums = 0 if argnums is None else argnums
    return _jacrev(fun=func,
                   argnums=argnums,
                   holomorphic=holomorphic,
                   allow_int=allow_int,
                   has_aux=has_aux,
                   return_value=return_value)
  else:
    _argnums, _ = tree_flatten(argnums)
    _argnums = tuple(a + 2 for a in _argnums)
    if argnums is None and len(grad_vars) == 0:
      raise errors.MathError('We detect no require to compute gradients because '
                             '"grad_vars" is None and "argnums" is also None. '
                             'Please provide one of them.')
    # computation
    grad_func = _cls_jacrev(func=func,
                            grad_vars=grad_vars,
                            dyn_vars=dyn_vars,
                            argnums=(0,) + _argnums,
                            has_aux=has_aux,
                            holomorphic=holomorphic,
                            allow_int=allow_int)

    call_func = _make_cls_call_func(grad_func=grad_func,
                                    grad_tree=grad_tree,
                                    grad_vars=grad_vars,
                                    dyn_vars=dyn_vars,
                                    argnums=argnums,
                                    return_value=return_value,
                                    has_aux=has_aux)
    return call_func


jacobian = jacrev


def _jacfwd(fun, argnums=0, holomorphic=False, has_aux=False, return_value=False):
  _check_callable(fun)
  if has_aux and jax.__version__ < '0.2.28':
    raise NotImplementedError(f'"has_aux" only supported in jax>=0.2.28, but we detect '
                              f'the current jax version is {jax.__version__}')

  def jacfun(*args, **kwargs):
    f = linear_util.wrap_init(fun, kwargs)
    f_partial, dyn_args = argnums_partial(f, argnums, args, require_static_args_hashable=False)
    tree_map(partial(_check_input_dtype_jacfwd, holomorphic), dyn_args)
    if has_aux:
      pushfwd = partial(_jvp, f_partial, dyn_args, has_aux=True)
      y, jac, aux = vmap(pushfwd, out_axes=(None, -1, None))(_std_basis(dyn_args))
    else:
      pushfwd = partial(_jvp, f_partial, dyn_args)
      y, jac = vmap(pushfwd, out_axes=(None, -1))(_std_basis(dyn_args))
    tree_map(partial(_check_output_dtype_jacfwd, holomorphic), y)
    example_args = dyn_args[0] if isinstance(argnums, int) else dyn_args
    jac = tree_map(partial(_unravel_array_into_pytree, example_args, -1, is_leaf=_isleaf), jac, is_leaf=_isleaf)
    if return_value:
      return (jac, y, aux) if has_aux else (jac, y)
    else:
      return (jac, aux) if has_aux else jac

  return jacfun


def _cls_jacfwd(func, grad_vars, dyn_vars, argnums, holomorphic=False, has_aux=False):
  # parameters
  assert isinstance(dyn_vars, (tuple, list))  # tuple/list of JaxArray
  assert isinstance(grad_vars, (tuple, list))  # tuple/list of JaxArray
  assert isinstance(argnums, (tuple, list))  # tuple/list of int

  # final functions
  if has_aux:
    @partial(_jacfwd,
             argnums=argnums, holomorphic=holomorphic, has_aux=True)
    def grad_func(grad_values, dyn_values, *args, **kwargs):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      for v, d in zip(grad_vars, grad_values): v.value = d
      # outputs: [0] is the value for gradient,
      #          [1] is other values for return
      outputs = func(*args, **kwargs)
      output = outputs[0].value if isinstance(outputs[0], JaxArray) else outputs[0]
      return output, (outputs, [v.value for v in grad_vars], [v.value for v in dyn_vars])

  else:
    @partial(_jacfwd,
             argnums=argnums, holomorphic=holomorphic, has_aux=True)
    def grad_func(grad_values, dyn_values, *args, **kwargs):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      for v, d in zip(grad_vars, grad_values): v.value = d
      outputs = func(*args, **kwargs)
      output = outputs.value if isinstance(outputs, JaxArray) else outputs
      return output, (outputs, [v.value for v in grad_vars], [v.value for v in dyn_vars])

  return grad_func


def jacfwd(func, grad_vars=None, dyn_vars=None, argnums=None, holomorphic=False,
           has_aux=None, return_value=False):
  """Extending automatic Jacobian (forward-mode) of ``func`` to classes.

  This function extends the JAX official ``jacfwd`` to make automatic jacobian
  computation on functions and class functions. Moreover, it supports returning
  value ("return_value") and returning auxiliary data ("has_aux").

  Same as `brainpy.math.grad <./brainpy.math.autograd.grad.html>`_, the returns are
  different for different argument settings in ``brainpy.math.jacfwd``.

  1. When "grad_vars" is None
    - "has_aux=False" + "return_value=False" => ``arg_grads``.
    - "has_aux=True" + "return_value=False" => ``(arg_grads, aux_data)``.
    - "has_aux=False" + "return_value=True" => ``(arg_grads, loss_value)``.
    - "has_aux=True" + "return_value=True" => ``(arg_grads, loss_value, aux_data)``.
  2. When "grad_vars" is not None and "argnums" is None
    - "has_aux=False" + "return_value=False" => ``var_grads``.
    - "has_aux=True" + "return_value=False" => ``(var_grads, aux_data)``.
    - "has_aux=False" + "return_value=True" => ``(var_grads, loss_value)``.
    - "has_aux=True" + "return_value=True" => ``(var_grads, loss_value, aux_data)``.
  3. When "grad_vars" is not None and "argnums" is not None
    - "has_aux=False" + "return_value=False" => ``(var_grads, arg_grads)``.
    - "has_aux=True" + "return_value=False" => ``((var_grads, arg_grads), aux_data)``.
    - "has_aux=False" + "return_value=True" => ``((var_grads, arg_grads), loss_value)``.
    - "has_aux=True" + "return_value=True" => ``((var_grads, arg_grads), loss_value, aux_data)``.

  Parameters
  ----------
  func: Function whose Jacobian is to be computed.
  dyn_vars : optional, JaxArray, sequence of JaxArray, dict
    The dynamically changed variables used in ``func``.
  grad_vars : optional, JaxArray, sequence of JaxArray, dict
    The variables in ``func`` to take their gradients.
  has_aux: optional, bool
    Indicates whether ``fun`` returns a pair where the
    first element is considered the output of the mathematical function to be
    differentiated and the second element is auxiliary data. Default False.
  return_value : bool
    Whether return the loss value.
  argnums: Optional, integer or sequence of integers. Specifies which
    positional argument(s) to differentiate with respect to (default ``0``).
  holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
    holomorphic. Default False.
  """
  dyn_vars, grad_vars, grad_tree = _grad_checking(func, dyn_vars, grad_vars)
  has_aux = False if has_aux is None else has_aux

  if (len(dyn_vars) == 0) and (len(grad_vars) == 0):
    argnums = 0 if argnums is None else argnums
    return _jacfwd(fun=func,
                   argnums=argnums,
                   holomorphic=holomorphic,
                   has_aux=has_aux,
                   return_value=return_value)
  else:
    _argnums, _ = tree_flatten(argnums)
    _argnums = tuple(a + 2 for a in _argnums)
    if argnums is None and len(grad_vars) == 0:
      raise errors.MathError('We detect no require to compute gradients because '
                             '"grad_vars" is None and "argnums" is also None. '
                             'Please provide one of them.')
    # computation
    grad_func = _cls_jacfwd(func=func,
                            grad_vars=grad_vars,
                            dyn_vars=dyn_vars,
                            argnums=(0,) + _argnums,
                            has_aux=has_aux,
                            holomorphic=holomorphic)

    call_func = _make_cls_call_func(grad_func=grad_func,
                                    grad_tree=grad_tree,
                                    grad_vars=grad_vars,
                                    dyn_vars=dyn_vars,
                                    argnums=argnums,
                                    return_value=return_value,
                                    has_aux=has_aux)

    return call_func


def hessian(func, dyn_vars=None, grad_vars=None, argnums=None, holomorphic=False, return_value=False):
  """Hessian of ``func`` as a dense array.

  Parameters
  ----------
  func : callable, function
    Function whose Hessian is to be computed.  Its arguments at positions
    specified by ``argnums`` should be arrays, scalars, or standard Python
    containers thereof. It should return arrays, scalars, or standard Python
    containers thereof.
  dyn_vars : optional, ArrayCollector, sequence of JaxArray
    The dynamical changed variables.
  grad_vars : optional, ArrayCollector, sequence of JaxArray
    The variables required to compute their gradients.
  argnums: Optional, integer or sequence of integers
    Specifies which positional argument(s) to differentiate with respect to (default ``0``).
  holomorphic : bool
    Indicates whether ``fun`` is promised to be holomorphic. Default False.
  return_value : bool
    Whether return the hessian values.
  """
  dyn_vars, grad_vars, grad_tree = _grad_checking(func, dyn_vars, grad_vars)
  argnums = 0 if argnums is None else argnums

  if (len(dyn_vars) == 0) and (len(grad_vars) == 0) and (not return_value):
    return jax.hessian(func, argnums=argnums, holomorphic=holomorphic)
  else:
    return jacfwd(jacrev(func,
                         dyn_vars=dyn_vars,
                         grad_vars=grad_vars,
                         argnums=argnums,
                         holomorphic=holomorphic),
                  dyn_vars=dyn_vars,
                  grad_vars=grad_vars,
                  argnums=argnums,
                  holomorphic=holomorphic,
                  return_value=return_value)


def _vector_grad(func, argnums=0, return_value=False, has_aux=False):
  _check_callable(func)

  def grad_fun(*args, **kwargs):
    f = linear_util.wrap_init(func, kwargs)
    f_partial, dyn_args = argnums_partial(f, argnums, args, require_static_args_hashable=False)
    if has_aux:
      y, vjp_fn, aux = _vjp(f_partial, *dyn_args, has_aux=True)
    else:
      y, vjp_fn = _vjp(f_partial, *dyn_args, has_aux=False)
    leaves, tree = tree_flatten(y)
    tangents = tree_unflatten(tree, [jnp.ones_like(l) for l in leaves])
    grads = vjp_fn(tangents)
    if isinstance(argnums, int):
      grads = grads[0]
    if has_aux:
      return (grads, y, aux) if return_value else (grads, aux)
    else:
      return (grads, y) if return_value else grads

  return grad_fun


def _cls_vector_grad(func, grad_vars, dyn_vars, argnums, has_aux=False):
  # parameters
  assert isinstance(dyn_vars, (tuple, list))  # tuple/list of JaxArray
  assert isinstance(grad_vars, (tuple, list))  # tuple/list of JaxArray
  assert isinstance(argnums, (tuple, list))  # tuple/list of int

  # final functions
  if has_aux:
    @partial(_vector_grad, argnums=argnums, has_aux=True)
    def grad_func(grad_values, dyn_values, *args, **kwargs):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      for v, d in zip(grad_vars, grad_values): v.value = d
      outputs = func(*args, **kwargs)
      output = outputs[0].value if isinstance(outputs[0], JaxArray) else outputs[0]
      return output, (outputs, [v.value for v in grad_vars], [v.value for v in dyn_vars])

  else:
    @partial(_vector_grad, argnums=argnums, has_aux=True)
    def grad_func(grad_values, dyn_values, *args, **kwargs):
      for v, d in zip(dyn_vars, dyn_values): v.value = d
      for v, d in zip(grad_vars, grad_values): v.value = d
      outputs = func(*args, **kwargs)
      output = outputs.value if isinstance(outputs, JaxArray) else outputs
      return output, (outputs, [v.value for v in grad_vars], [v.value for v in dyn_vars])

  return grad_func


def vector_grad(func, dyn_vars=None, grad_vars=None, argnums=None, return_value=False, has_aux=None):
  """Take vector-valued gradients for function ``func``.

  Same as `brainpy.math.grad <./brainpy.math.autograd.grad.html>`_,
  `brainpy.math.jacrev <./brainpy.math.autograd.jacrev.html>`_ and
  `brainpy.math.jacfwd <./brainpy.math.autograd.jacfwd.html>`_,
  the returns in this function are different for different argument settings.

  1. When "grad_vars" is None
    - "has_aux=False" + "return_value=False" => ``arg_grads``.
    - "has_aux=True" + "return_value=False" => ``(arg_grads, aux_data)``.
    - "has_aux=False" + "return_value=True" => ``(arg_grads, loss_value)``.
    - "has_aux=True" + "return_value=True" => ``(arg_grads, loss_value, aux_data)``.
  2. When "grad_vars" is not None and "argnums" is None
    - "has_aux=False" + "return_value=False" => ``var_grads``.
    - "has_aux=True" + "return_value=False" => ``(var_grads, aux_data)``.
    - "has_aux=False" + "return_value=True" => ``(var_grads, loss_value)``.
    - "has_aux=True" + "return_value=True" => ``(var_grads, loss_value, aux_data)``.
  3. When "grad_vars" is not None and "argnums" is not None
    - "has_aux=False" + "return_value=False" => ``(var_grads, arg_grads)``.
    - "has_aux=True" + "return_value=False" => ``((var_grads, arg_grads), aux_data)``.
    - "has_aux=False" + "return_value=True" => ``((var_grads, arg_grads), loss_value)``.
    - "has_aux=True" + "return_value=True" => ``((var_grads, arg_grads), loss_value, aux_data)``.


  Parameters
  ----------
  func: Function whose Jacobian is to be computed.
  dyn_vars : optional, JaxArray, sequence of JaxArray, dict
    The dynamically changed variables used in ``func``.
  grad_vars : optional, JaxArray, sequence of JaxArray, dict
    The variables in ``func`` to take their gradients.
  has_aux: optional, bool
    Indicates whether ``fun`` returns a pair where the
    first element is considered the output of the mathematical function to be
    differentiated and the second element is auxiliary data. Default False.
  return_value : bool
    Whether return the loss value.
  argnums: Optional, integer or sequence of integers. Specifies which
    positional argument(s) to differentiate with respect to (default ``0``).

  Returns
  -------
  func : callable
    The vector gradient function.
  """
  dyn_vars, grad_vars, grad_tree = _grad_checking(func, dyn_vars, grad_vars)
  has_aux = False if has_aux is None else has_aux

  if (len(dyn_vars) == 0) and (len(grad_vars) == 0):
    argnums = 0 if argnums is None else argnums
    call_func = _vector_grad(func=func,
                             argnums=argnums,
                             return_value=return_value,
                             has_aux=has_aux)

  else:
    _argnums, _ = tree_flatten(argnums)
    _argnums = tuple(a + 2 for a in _argnums)
    if argnums is None and len(grad_vars) == 0:
      raise errors.MathError('We detect no require to compute gradients because '
                             '"grad_vars" is None and "argnums" is also None. '
                             'Please provide one of them.')
    # computation
    grad_func = _cls_vector_grad(func=func,
                                 grad_vars=grad_vars,
                                 dyn_vars=dyn_vars,
                                 argnums=(0,) + _argnums,
                                 has_aux=has_aux)

    call_func = _make_cls_call_func(grad_func=grad_func,
                                    grad_tree=grad_tree,
                                    grad_vars=grad_vars,
                                    dyn_vars=dyn_vars,
                                    argnums=argnums,
                                    return_value=return_value,
                                    has_aux=has_aux)

  return call_func



# -*- coding: utf-8 -*-

import numpy as np
import jax.numpy as jnp
from jax import jit

from brainpy import math as bm

__all__ = [
  'raster_plot',
  'firing_rate',
]


def raster_plot(sp_matrix, times):
  """Get spike raster plot which displays the spiking activity
  of a group of neurons over time.

  Parameters
  ----------
  sp_matrix : bnp.ndarray
      The matrix which record spiking activities.
  times : bnp.ndarray
      The time steps.

  Returns
  -------
  raster_plot : tuple
      Include (neuron index, spike time).
  """
  sp_matrix = bm.as_numpy(sp_matrix)
  times = np.asarray(times)
  elements = np.where(sp_matrix > 0.)
  index = elements[1]
  time = times[elements[0]]
  return index, time


@jit
def _firing_rate(sp_matrix, window):
  sp_matrix = bm.as_device_array(sp_matrix)
  rate = jnp.sum(sp_matrix, axis=1) / sp_matrix.shape[1]
  return jnp.convolve(rate, window, mode='same')


def firing_rate(sp_matrix, width, dt=None, numpy=True):
  r"""Calculate the mean firing rate over in a neuron group.

  This method is adopted from Brian2.

  The firing rate in trial :math:`k` is the spike count :math:`n_{k}^{sp}`
  in an interval of duration :math:`T` divided by :math:`T`:

  .. math::

      v_k = {n_k^{sp} \over T}

  Parameters
  ----------
  sp_matrix : math.JaxArray, np.ndarray
    The spike matrix which record spiking activities.
  width : int, float
    The width of the ``window`` in millisecond.
  dt : float, optional
    The sample rate.

  Returns
  -------
  rate : numpy.ndarray
      The population rate in Hz, smoothed with the given window.
  """
  dt = bm.get_dt() if (dt is None) else dt
  width1 = int(width / 2 / dt) * 2 + 1
  window = jnp.ones(width1) * 1000 / width
  fr = _firing_rate(sp_matrix, window)
  return bm.as_numpy(fr) if numpy else fr




# -*- coding: utf-8 -*-

import ast
import inspect
from collections import OrderedDict

from brainpy import errors, tools

__all__ = [
  'DiffEqReader',
  'separate_variables',
]


class DiffEqReader(ast.NodeVisitor):
  """Read the code lines which defines the logic of a differential equation system.

  Currently, DiffEqReader cannot handle the for loop, and if-else condition.
  Also, it do not assign values by a functional call. Like this:

  .. code-block:: python

      func(a, b, c)

  Instead, you should code like:

  .. code-block:: python

      c = func(a, b)

  Therefore, this class only has minimum power to analyze differential
  equations. For example, this class may help to automatically find out
  the linear part of a differential equation, thus forming the
  Exponential Euler numerical methods.
  """

  return_error_msg = '''Cannot analyze differential equation with expression return. Like:

def df(v, t):
    return -v + 1.

We only support return variables. Therefore, the above should be coded as:

def df(v, t):
    dv = -v + 1.
    return dv
'''

  def __init__(self):
    self.code_lines = []  # list of str
    self.variables = []  # list of list
    self.returns = []  # list of str
    self.rights = []  # list of str

  @staticmethod
  def visit_container(nodes):
    variables = []
    for var in nodes:
      if isinstance(var, (ast.List, ast.Tuple)):
        variables.extend(DiffEqReader.visit_container(var.elts))
      elif isinstance(var, ast.Name):
        variables.extend(var.id)
      else:
        raise ValueError(f'Unknown target type: {var}')
    return variables

  def visit_Assign(self, node):
    variables = []
    for target in node.targets:
      if isinstance(target, (ast.List, ast.Tuple)):
        variables.extend(self.visit_container(target.elts))
      elif isinstance(target, ast.Name):
        variables.append(target.id)
      else:
        raise ValueError(f'Unknown target type: {target}')
    self.variables.append(variables)
    self.code_lines.append(tools.ast2code(ast.fix_missing_locations(node)))
    self.rights.append(tools.ast2code(ast.fix_missing_locations(node.value)))
    return node

  def visit_AugAssign(self, node):
    var = node.target.id
    self.variables.append(var)
    expr = tools.ast2code(ast.fix_missing_locations(node))
    self.code_lines.append(expr)
    self.rights.append(tools.ast2code(ast.fix_missing_locations(node.value)))
    return node

  def visit_Return(self, node):
    if isinstance(node.value, ast.Name):
      self.returns.append(node.value.id)
    elif isinstance(node.value, (ast.Tuple, ast.List)):
      for var in node.value.elts:
        if not isinstance(var, ast.Name):
          raise errors.DiffEqError(self.return_error_msg)
        self.returns.append(var.id)
    else:
      raise errors.DiffEqError(self.return_error_msg)
    return node

  def visit_AnnAssign(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze an assignment with a type annotation.')

  def visit_If(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze "if-else" conditions in differential equation.')

  def visit_IfExp(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze "if-else" conditions in differential equation.')

  def visit_For(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze "for" loops in differential equation.')

  def visit_While(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze "while" loops in differential equation.')

  def visit_Try(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze "try" handler in differential equation.')

  def visit_With(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze "with" block in differential equation.')

  def visit_Raise(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze "raise" statement in differential equation.')

  def visit_Delete(self, node):
    raise errors.DiffEqError(f'Currently, {self.__class__.__name__} do not support to '
                             f'analyze "del" operation in differential equation.')


def separate_variables(func_or_code):
  """Separate the expressions in a differential equation for each variable.

  For example, take the HH neuron model as an example:

  >>> eq_code = '''
  >>> def derivative(V, m, h, n, t, C, gNa, ENa, gK, EK, gL, EL, Iext):
  >>>     alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
  >>>     beta = 4.0 * bp.math.exp(-(V + 65) / 18)
  >>>     dmdt = alpha * (1 - m) - beta * m
  >>>
  >>>     alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
  >>>     beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
  >>>     dhdt = alpha * (1 - h) - beta * h
  >>>
  >>>     alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
  >>>     beta = 0.125 * bp.math.exp(-(V + 65) / 80)
  >>>     dndt = alpha * (1 - n) - beta * n
  >>>
  >>>     I_Na = (gNa * m ** 3.0 * h) * (V - ENa)
  >>>     I_K = (gK * n ** 4.0) * (V - EK)
  >>>     I_leak = gL * (V - EL)
  >>>     dVdt = (- I_Na - I_K - I_leak + Iext) / C
  >>>
  >>>     return dVdt, dmdt, dhdt, dndt
  >>> '''
  >>> separate_variables(eq_code)
  {'code_lines_for_returns': {'dVdt': ['I_Na = gNa * m ** 3.0 * h * (V - ENa)\n',
                                       'I_K = gK * n ** 4.0 * (V - EK)\n',
                                       'I_leak = gL * (V - EL)\n',
                                       'dVdt = (-I_Na - I_K - I_leak + Iext) / C\n'],
                              'dhdt': ['alpha = 0.07 * bp.math.exp(-(V + 65) / 20.0)\n',
                                       'beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))\n',
                                       'dhdt = alpha * (1 - h) - beta * h\n'],
                              'dmdt': ['alpha = 0.1 * (V + 40) / (1 - '
                                       'bp.math.exp(-(V + 40) / 10))\n',
                                       'beta = 4.0 * bp.math.exp(-(V + 65) / 18)\n',
                                       'dmdt = alpha * (1 - m) - beta * m\n'],
                              'dndt': ['alpha = 0.01 * (V + 55) / (1 - '
                                       'bp.math.exp(-(V + 55) / 10))\n',
                                       'beta = 0.125 * bp.math.exp(-(V + 65) / 80)\n',
                                       'dndt = alpha * (1 - n) - beta * n\n']},
   'expressions_for_returns': {'dVdt': ['gNa * m ** 3.0 * h * (V - ENa)',
                                        'gK * n ** 4.0 * (V - EK)',
                                        'gL * (V - EL)',
                                        '(-I_Na - I_K - I_leak + Iext) / C'],
                               'dhdt': ['0.07 * bp.math.exp(-(V + 65) / 20.0)',
                                        '1 / (1 + bp.math.exp(-(V + 35) / 10))',
                                        'alpha * (1 - h) - beta * h'],
                               'dmdt': ['0.1 * (V + 40) / (1 - '
                                        'bp.math.exp(-(V + 40) / 10))',
                                        '4.0 * bp.math.exp(-(V + 65) / 18)',
                                        'alpha * (1 - m) - beta * m'],
                               'dndt': ['0.01 * (V + 55) / (1 - '
                                        'bp.math.exp(-(V + 55) / 10))',
                                        '0.125 * bp.math.exp(-(V + 65) / 80)',
                                        'alpha * (1 - n) - beta * n']},
   'variables_for_returns': {'dVdt': [['I_Na'], ['I_K'], ['I_leak'], ['dVdt']],
                             'dhdt': [['alpha'], ['beta'], ['dhdt']],
                             'dmdt': [['alpha'], ['beta'], ['dmdt']],
                             'dndt': [['alpha'], ['beta'], ['dndt']]}}

  Parameters
  ----------
  func_or_code : callable, str
      The callable function or the function code.

  Returns
  -------
  anlysis : dict
      The expressions for each return variable.
  """
  if callable(func_or_code):
    if tools.is_lambda_function(func_or_code):
      raise errors.AnalyzerError(f'Cannot analyze lambda function: {func_or_code}.')
    func_or_code = tools.deindent(inspect.getsource(func_or_code))
  assert isinstance(func_or_code, str)
  analyser = DiffEqReader()
  analyser.visit(ast.parse(func_or_code))

  returns = analyser.returns
  variables = analyser.variables
  right_exprs = analyser.rights
  code_lines = analyser.code_lines

  return_requires = OrderedDict([(r, set(tools.get_identifiers(r))) for r in returns])
  code_lines_for_returns = OrderedDict([(r, []) for r in returns])
  variables_for_returns = OrderedDict([(r, []) for r in returns])
  expressions_for_returns = OrderedDict([(r, []) for r in returns])

  length = len(variables)
  reverse_ids = list(reversed([i - length for i in range(length)]))
  for r in code_lines_for_returns.keys():
    for rid in reverse_ids:
      dep = []
      for v in variables[rid]:
        if v in return_requires[r]:
          dep.append(v)
      if len(dep):
        code_lines_for_returns[r].append(code_lines[rid])
        variables_for_returns[r].append(variables[rid])
        expr = right_exprs[rid]
        expressions_for_returns[r].append(expr)
        for d in dep:
          return_requires[r].remove(d)
        return_requires[r].update(tools.get_identifiers(expr))
  for r in list(code_lines_for_returns.keys()):
    code_lines_for_returns[r] = code_lines_for_returns[r][::-1]
    variables_for_returns[r] = variables_for_returns[r][::-1]
    expressions_for_returns[r] = expressions_for_returns[r][::-1]

  analysis = tools.DictPlus(
    code_lines_for_returns=code_lines_for_returns,
    variables_for_returns=variables_for_returns,
    expressions_for_returns=expressions_for_returns,
  )
  return analysis



# -*- coding: utf-8 -*-

import time

import jax.numpy as jnp
import numpy as np
import tqdm.auto
from jax.experimental.host_callback import id_tap

from brainpy.base.base import TensorCollector
from brainpy import math as bm
from brainpy.dyn import utils
from brainpy.dyn.base import DynamicalSystem
from brainpy.errors import RunningError
from brainpy.running.runner import Runner

__all__ = [
  'DSRunner', 'ReportRunner', 'StructRunner',
]


class DSRunner(Runner):
  """The runner for dynamical systems.

  Parameters
  ----------
  target : DynamicalSystem
    The target model to run.
  inputs : list, tuple
    The inputs for the target DynamicalSystem. It should be the format
    of `[(target, value, [type, operation])]`, where `target` is the
    input target, `value` is the input value, `type` is the input type
    (such as "fix" or "iter"), `operation` is the operation for inputs
    (such as "+", "-", "*", "/", "=").

    - ``target``: should be a string. Can be specified by the *absolute access* or *relative access*.
    - ``value``: should be a scalar, vector, matrix, iterable function or objects.
    - ``type``: should be a string. "fix" means the input `value` is a constant. "iter" means the
      input `value` can be changed over time.
    - ``operation``: should be a string, support `+`, `-`, `*`, `/`, `=`.
    - Also, if you want to specify multiple inputs, just give multiple ``(target, value, [type, operation])``,
      for example ``[(target1, value1), (target2, value2)]``.
  """

  def __init__(self, target: DynamicalSystem, inputs=(), dt=None, **kwargs):
    super(DSRunner, self).__init__(target=target, **kwargs)

    # parameters
    dt = bm.get_dt() if dt is None else dt
    if not isinstance(dt, (int, float)):
      raise RunningError(f'"dt" must be scalar, but got {dt}')
    self.dt = dt
    if not isinstance(target, DynamicalSystem):
      raise RunningError(f'"target" must be an instance of {DynamicalSystem.__name__}, '
                         f'but we got {type(target)}: {target}')

    # Build the monitor function
    self._monitor_step = self.build_monitors()

    # whether has iterable input data
    self._has_iter_array = False  # default do not have iterable input array
    self._i = bm.Variable(bm.asarray([0]))

    # Build input function
    inputs = utils.check_and_format_inputs(host=target, inputs=inputs)
    self._input_step = self.build_inputs(inputs)

    # start simulation time
    self._start_t = None

    # JAX does not support iterator in fori_loop, scan, etc.
    #   https://github.com/google/jax/issues/3567
    # We use Variable i to index the current input data.
    if self._has_iter_array:  # must behind of "self.build_input()"
      self.dyn_vars.update({'_i': self._i})
    else:
      self._i = None

    # run function
    self._run_func = self.build_run_function()

  def build_inputs(self, inputs):
    fix_inputs = {'=': [], '+': [], '-': [], '*': [], '/': []}
    next_inputs = {'=': [], '+': [], '-': [], '*': [], '/': []}
    func_inputs = {'=': [], '+': [], '-': [], '*': [], '/': []}
    array_inputs = {'=': [], '+': [], '-': [], '*': [], '/': []}

    for target, key, value, type_, op in inputs:
      # variable
      variable = getattr(target, key)
      if not isinstance(variable, bm.Variable):
        raise RunningError(f'"{key}" in {target} is not a dynamically changed Variable, '
                           f'its value will not change, we think there is no need to '
                           f'give its input.')

      # input data
      if type_ == 'iter':
        if isinstance(value, (bm.ndarray, np.ndarray, jnp.ndarray)):
          array_inputs[op].append([variable, bm.asarray(value)])
          self._has_iter_array = True
        else:
          next_inputs[op].append([variable, iter(value)])
      elif type_ == 'func':
        func_inputs[op].append([variable, value])
      else:
        fix_inputs[op].append([variable, value])

    def _f_ops(ops, var, data):
      if ops == '=':
        var[:] = data
      elif ops == '+':
        var += data
      elif ops == '-':
        var -= data
      elif ops == '*':
        var *= data
      elif ops == '/':
        var /= data
      else:
        raise ValueError

    def func(_t, _dt):
      for ops, values in fix_inputs.items():
        for var, data in values:
          _f_ops(ops, var, data)
      for ops, values in array_inputs.items():
        for var, data in values:
          _f_ops(ops, var, data[self._i[0]])
      for ops, values in func_inputs.items():
        for var, data in values:
          _f_ops(ops, var, data(_t, _dt))
      for ops, values in next_inputs.items():
        for var, data in values:
          _f_ops(ops, var, next(data))
      if self._has_iter_array:
        self._i += 1

    return func

  def build_monitors(self):
    monitors = utils.check_and_format_monitors(host=self.target, mon=self.mon)

    return_with_idx = dict()
    return_without_idx = dict()
    for key, target, variable, idx, interval in monitors:
      if interval is not None:
        raise ValueError(f'Running with "{self.__class__.__name__}" does '
                         f'not support "interval" in the monitor.')
      data = target
      for k in variable.split('.'):
        data = getattr(data, k)
      if not isinstance(data, bm.Variable):
        raise RunningError(f'"{key}" in {target} is not a dynamically changed Variable, '
                           f'its value will not change, we think there is no need to '
                           f'monitor its trajectory.')
      if idx is None:
        return_without_idx[key] = data
      else:
        return_with_idx[key] = (data, bm.asarray(idx))

    def func(_t, _dt):
      res = {k: (v.flatten() if bm.ndim(v) > 1 else v.value)
             for k, v in return_without_idx.items()}
      res.update({k: (v.flatten()[idx] if bm.ndim(v) > 1 else v[idx])
                  for k, (v, idx) in return_with_idx.items()})
      return res

    return func

  def _run_one_step(self, _t):
    self._input_step(_t=_t, _dt=self.dt)
    self.target.update(_t=_t, _dt=self.dt)
    if self.progress_bar:
      id_tap(lambda *args: self._pbar.update(), ())
    return self._monitor_step(_t=_t, _dt=self.dt)

  def build_run_function(self):
    if self.jit:
      dyn_vars = TensorCollector()
      dyn_vars.update(self.dyn_vars)
      dyn_vars.update(self.target.vars().unique())
      f_run = bm.make_loop(self._run_one_step,
                           dyn_vars=dyn_vars,
                           has_return=True)
    else:
      def f_run(all_t):
        for i in range(all_t.shape[0]):
          mon = self._run_one_step(all_t[i])
          for k, v in mon.items():
            self.mon.item_contents[k].append(v)
        return None, {}
    return f_run

  def run(self, duration, start_t=None):
    return self.__call__(duration, start_t=start_t)

  def __call__(self, duration, start_t=None):
    """The running function.

    Parameters
    ----------
    duration : float, int, tuple, list
      The running duration.
    start_t : float, optional

    Returns
    -------
    running_time : float
      The total running time.
    """
    # time step
    if start_t is None:
      if self._start_t is None:
        start_t = 0.
      else:
        start_t = float(self._start_t)
    end_t = float(start_t + duration)
    # times
    times = np.arange(start_t, end_t, self.dt)
    # build monitor
    for key in self.mon.item_contents.keys():
      self.mon.item_contents[key] = []  # reshape the monitor items
    # running
    if self.progress_bar:
      self._pbar = tqdm.auto.tqdm(total=times.size)
      self._pbar.set_description(f"Running a duration of {round(float(duration), 3)} ({times.size} steps)",
                                 refresh=True)
    t0 = time.time()
    _, hists = self._run_func(times)
    running_time = time.time() - t0
    if self.progress_bar:
      self._pbar.close()
    # post-running
    if self.jit:
      self.mon.ts = times + self.dt
      for key in self.mon.item_names:
        self.mon.item_contents[key] = bm.asarray(hists[key])
    else:
      self.mon.ts = times + self.dt
      for key in self.mon.item_names:
        self.mon.item_contents[key] = bm.asarray(self.mon.item_contents[key])
    self._start_t = end_t
    if self.numpy_mon_after_run:
      self.mon.numpy()
    return running_time


class StructRunner(DSRunner):
  """The runner with the structural for-loop.

  .. deprecated:: 2.0.3
     Prefer the use of :py:class:`brainpy.dyn.DSRunner` for dynamical system running.
     This runner is deprecated since 2.0.3.
  """

  def __init__(self, target, *args, **kwargs):
    super(StructRunner, self).__init__(target, *args, **kwargs)


class ReportRunner(DSRunner):
  """The runner provides convenient interface for debugging.
  It is also able to report the running progress.

  .. deprecated:: 2.0.3
     Prefer the use of :py:class:`brainpy.dyn.DSRunner` for dynamical system running.
     This runner is deprecated since 2.0.3.

  Parameters
  ----------
  target : DynamicalSystem
    The target model to run.
  monitors : None, list of str, tuple of str, Monitor
    Variables to monitor.
  inputs : list, tuple
    The input settings.
  """

  def __init__(self, target, inputs=(), jit=False, dt=None, **kwargs):
    super(ReportRunner, self).__init__(target=target, inputs=inputs, dt=dt, jit=False, **kwargs)

    # Build the update function
    if jit:
      dyn_vars = TensorCollector()
      dyn_vars.update(self.dyn_vars)
      dyn_vars.update(self.target.vars().unique())
      self._update_step = bm.jit(self.target.update, dyn_vars=dyn_vars)
    else:
      self._update_step = self.target.update

  def _run_one_step(self, _t):
    self._input_step(_t, self.dt)
    self._update_step(_t, self.dt)
    if self.progress_bar:
      self._pbar.update()
    return self._monitor_step(_t, self.dt)

  def build_run_function(self):
    def f_run(all_t):
      for i in range(all_t.shape[0]):
        mon = self._run_one_step(all_t[i])
        for k, v in mon.items():
          self.mon.item_contents[k].append(v)
      return None, {}

    return f_run



# -*- coding: utf-8 -*-

import jax.numpy as jnp

from brainpy.base.base import Base
from brainpy.errors import MathError
from brainpy.math.jaxarray import Variable

__all__ = [
  # schedulers
  'make_schedule',
  'Scheduler',
  'Constant',
  'ExponentialDecay',
  'InverseTimeDecay',
  'PolynomialDecay',
  'PiecewiseConstant',
]

# learning rate schedules #
# ----------------------- #


def make_schedule(scalar_or_schedule):
  if isinstance(scalar_or_schedule, Scheduler):
    return scalar_or_schedule
  elif isinstance(scalar_or_schedule, (int, float)):
    return Constant(scalar_or_schedule)
  else:
    raise TypeError(type(scalar_or_schedule))


class Scheduler(Base):
  """The learning rate scheduler."""

  def __init__(self, lr):
    super(Scheduler, self).__init__()

    assert isinstance(lr, (float, int))
    self.lr = lr
    self.step = Variable(jnp.array([0]))

  def update(self):
    self.step += 1

  def __call__(self, i=None):
    raise NotImplementedError

  def __repr__(self):
    return f'{self.__class__.__name__}({self.lr})'


class Constant(Scheduler):
  def __call__(self, i=None):
    return self.lr


class ExponentialDecay(Scheduler):
  def __init__(self, lr, decay_steps, decay_rate):
    super(ExponentialDecay, self).__init__(lr=lr)
    self.decay_steps = decay_steps
    self.decay_rate = decay_rate

  def __call__(self, i=None):
    i = self.step[0] if i is None else i
    return self.lr * self.decay_rate ** (i / self.decay_steps)

  def __repr__(self):
    return (f'{self.__class__.__name__}({self.lr}, '
            f'decay_steps={self.decay_steps}, '
            f'decay_rate={self.decay_rate})')


class InverseTimeDecay(ExponentialDecay):
  def __init__(self, lr, decay_steps, decay_rate, staircase=False):
    super(InverseTimeDecay, self).__init__(lr, decay_steps, decay_rate)
    self.staircase = staircase

  def __call__(self, i=None):
    i = self.step[0] if i is None else i
    if self.staircase:
      return self.lr / (1 + self.decay_rate * jnp.floor(i / self.decay_steps))
    else:
      return self.lr / (1 + self.decay_rate * i / self.decay_steps)

  def __repr__(self):
    return f'{self.__class__.__name__}({self.lr}, staircase={self.staircase})'


class PolynomialDecay(Scheduler):
  def __init__(self, lr, decay_steps, final_lr, power=1.0):
    super(PolynomialDecay, self).__init__(lr)
    self.decay_steps = decay_steps
    self.final_lr = final_lr
    self.power = power

  def __call__(self, i=None):
    i = self.step[0] if i is None else i
    i = jnp.minimum(i, self.decay_steps)
    step_mult = (1 - i / self.decay_steps) ** self.power
    return step_mult * (self.lr - self.final_lr) + self.final_lr

  def __repr__(self):
    return (f'{self.__class__.__name__}({self.lr}, '
            f'decay_steps={self.decay_steps}, '
            f'final_lr={self.final_lr}, '
            f'power={self.power})')


class PiecewiseConstant(Scheduler):
  def __init__(self, boundaries, values):
    super(PiecewiseConstant, self).__init__(0.)

    boundaries = jnp.array(boundaries)
    values = jnp.array(values)
    if not boundaries.ndim == values.ndim == 1:
      raise MathError("boundaries and values must be sequences")
    if not boundaries.shape[0] == values.shape[0] - 1:
      raise MathError("boundaries length must be one shorter than values length")
    self.boundaries = boundaries
    self.values = values

  def __call__(self, i=None):
    i = self.step[0] if i is None else i
    return self.values[jnp.sum(i > self.boundaries)]



# -*- coding: utf-8 -*-

"""
The parallel compilation tools for JAX backend.

1. Vectorize compilation is implemented by the 'vmap()' function
2. Parallel compilation is implemented by the 'pmap()' function

"""


import functools

import jax
import jax.numpy as jnp
import numpy as np
from jax.interpreters.partial_eval import DynamicJaxprTracer
from jax.interpreters.partial_eval import JaxprTracer
from jax.interpreters.pxla import ShardedDeviceArray

try:
  from jax.errors import UnexpectedTracerError
except ImportError:
  from jax.core import UnexpectedTracerError

from brainpy import errors
from brainpy.base.base import Base
from brainpy.base.collector import TensorCollector
from brainpy.math.random import RandomState
from brainpy.math.jaxarray import JaxArray
from brainpy.tools.codes import change_func_name

__all__ = [
  'vmap',
  'pmap',
]


def _make_vmap(func, nonbatched_vars, batched_vars, in_axes, out_axes,
               batch_idx, axis_name, f_name=None):
  @functools.partial(jax.vmap, in_axes=in_axes, out_axes=out_axes, axis_name=axis_name)
  def vmapped_func(nonbatched_data, batched_data, *args, **kwargs):
    nonbatched_vars.assign(nonbatched_data)
    batched_vars.assign(batched_data)
    out = func(*args, **kwargs)
    nonbatched_changes = nonbatched_vars.dict()
    batched_changes = batched_vars.dict()
    return nonbatched_changes, batched_changes, out

  def call(*args, **kwargs):
    n = args[batch_idx[0]].shape[batch_idx[1]]
    nonbatched_data = nonbatched_vars.dict()
    batched_data = {key: val.split_keys(n) for key, val in batched_vars.items()}
    try:
      out, dyn_changes, rand_changes = vmapped_func(nonbatched_data, batched_data, *args, **kwargs)
    except UnexpectedTracerError as e:
      nonbatched_vars.assign(nonbatched_data)
      batched_vars.assign(batched_data)
      raise errors.JaxTracerError() from e
    # for key, v in dyn_changes.items():
    #   dyn_vars[key] = reduce_func(v)
    # for key, v in rand_changes.items():
    #   rand_vars[key] = reduce_func(v)
    return out

  return change_func_name(name=f_name, f=call) if f_name else call


def vmap(func, dyn_vars=None, batched_vars=None,
         in_axes=0, out_axes=0, axis_name=None,
         reduce_func=None, auto_infer=False):
  """Vectorization compilation for class objects.

  Vectorized compile a function or a module to run in parallel on a single device.

  Examples
  --------

  Parameters
  ----------
  func : Base, function, callable
    The function or the module to compile.
  dyn_vars : dict, sequence
  batched_vars : dict
  in_axes : optional, int, sequence of int
    Specify which input array axes to map over. If each positional argument to
    ``obj_or_func`` is an array, then ``in_axes`` can be an integer, a None,
    or a tuple of integers and Nones with length equal to the number of
    positional arguments to ``obj_or_func``. An integer or ``None``
    indicates which array axis to map over for all arguments (with ``None``
    indicating not to map any axis), and a tuple indicates which axis to map
    for each corresponding positional argument. Axis integers must be in the
    range ``[-ndim, ndim)`` for each array, where ``ndim`` is the number of
    dimensions (axes) of the corresponding input array.

    If the positional arguments to ``obj_or_func`` are container types, the
    corresponding element of ``in_axes`` can itself be a matching container,
    so that distinct array axes can be mapped for different container
    elements. ``in_axes`` must be a container tree prefix of the positional
    argument tuple passed to ``obj_or_func``.

    At least one positional argument must have ``in_axes`` not None. The sizes
    of the mapped input axes for all mapped positional arguments must all be
    equal.

    Arguments passed as keywords are always mapped over their leading axis
    (i.e. axis index 0).
  out_axes : optional, int, tuple/list/dict
    Indicate where the mapped axis should appear in the output. All outputs
    with a mapped axis must have a non-None ``out_axes`` specification. Axis
    integers must be in the range ``[-ndim, ndim)`` for each output array,
    where ``ndim`` is the number of dimensions (axes) of the array returned
    by the :func:`vmap`-ed function, which is one more than the number of
    dimensions (axes) of the corresponding array returned by ``obj_or_func``.
  axis_name : optional

  Returns
  -------
  obj_or_func : Any
    Batched/vectorized version of ``obj_or_func`` with arguments that correspond to
    those of ``obj_or_func``, but with extra array axes at positions indicated by
    ``in_axes``, and a return value that corresponds to that of ``obj_or_func``, but
    with extra array axes at positions indicated by ``out_axes``.

  """
  # if isinstance(func, DynamicalSystem):
  #   if len(func.steps):  # DynamicalSystem has step functions
  #
  #     # dynamical variables
  #     dyn_vars = (dyn_vars or func.vars().unique())
  #     dyn_vars, rand_vars = TensorCollector(), TensorCollector()
  #     for key, val in dyn_vars.items():
  #       if isinstance(val, RandomState):
  #         rand_vars[key] = val
  #       else:
  #         dyn_vars[key] = val
  #
  #     # in axes
  #     if in_axes is None:
  #       in_axes = {key: (None, 0) for key in func.steps.keys()}
  #     elif isinstance(in_axes, int):
  #       in_axes = {key: (None, 0, in_axes) for key in func.steps.keys()}
  #     elif isinstance(in_axes, (tuple, list)):
  #       in_axes = {key: (None, 0) + tuple(in_axes) for key in func.steps.keys()}
  #     elif isinstance(in_axes, dict):
  #       keys = list(func.steps.keys())
  #       if keys[0] not in in_axes:
  #         in_axes = {key: (None, 0, in_axes) for key in keys}
  #       else:
  #         in_axes = {key: (None, 0) + tuple(in_axes[key]) for key in keys}
  #     assert isinstance(in_axes, dict)
  #
  #     # batch size index
  #     batch_idx = {}
  #     for key, axes in in_axes.items():
  #       for i, axis in enumerate(axes[2:]):
  #         if axis is not None:
  #           batch_idx[key] = (i, axis)
  #           break
  #       else:
  #         raise ValueError(f'Found no batch axis: {axes}.')
  #
  #     # out axes
  #     if out_axes is None:
  #       out_axes = {key: 0 for key in func.steps.keys()}
  #     elif isinstance(out_axes, int):
  #       out_axes = {key: out_axes for key in func.steps.keys()}
  #     elif isinstance(out_axes, (tuple, list)):
  #       out_axes = {key: tuple(out_axes) + (0, 0) for key in func.steps.keys()}
  #     elif isinstance(out_axes, dict):
  #       keys = list(func.steps.keys())
  #       if keys[0] not in out_axes:
  #         out_axes = {key: (out_axes, 0, 0) for key in keys}
  #       else:
  #         out_axes = {key: tuple(out_axes[key]) + (0, 0) for key in keys}
  #     assert isinstance(out_axes, dict)
  #
  #     # reduce_func
  #     if reduce_func is None:
  #       reduce_func = lambda x: x.mean(axis=0)
  #
  #     # vectorized map functions
  #     for key in func.steps.keys():
  #       func.steps[key] = _make_vmap(func=func.steps[key],
  #                                    dyn_vars=dyn_vars,
  #                                    rand_vars=rand_vars,
  #                                    in_axes=in_axes[key],
  #                                    out_axes=out_axes[key],
  #                                    axis_name=axis_name,
  #                                    batch_idx=batch_idx[key],
  #                                    reduce_func=reduce_func,
  #                                    f_name=key)
  #
  #     return func

  if callable(func):
    if auto_infer:
      if dyn_vars is not None:
        dyn_vars = dyn_vars
      elif isinstance(func, Base):  # Base has '__call__()' implementation
        dyn_vars = func.vars().unique()
      elif hasattr(func, '__self__'):
        if isinstance(func.__self__, Base):
          dyn_vars = func.__self__.vars().unique()

    if dyn_vars is None:
      return jax.vmap(func,
                      in_axes=in_axes,
                      out_axes=out_axes,
                      axis_name=axis_name)

    else:
      if isinstance(dyn_vars, JaxArray):
        dyn_vars = [dyn_vars]
      if isinstance(dyn_vars, (tuple, list)):
        dyn_vars = {f'_vmap_v{i}': v for i, v in enumerate(dyn_vars)}
      assert isinstance(dyn_vars, dict)

      # dynamical variables
      _dyn_vars, _rand_vars = TensorCollector(), TensorCollector()
      for key, val in dyn_vars.items():
        if isinstance(val, RandomState):
          _rand_vars[key] = val
        else:
          _dyn_vars[key] = val

      # in axes
      if in_axes is None:
        in_axes = (None, 0)
      elif isinstance(in_axes, (int, dict)):
        in_axes = (None, 0, in_axes)
      elif isinstance(in_axes, (tuple, list)):
        in_axes = (None, 0) + tuple(in_axes)
      assert isinstance(in_axes, (tuple, list))

      # batch size index
      batch_idx = {}
      for key, axes in batch_idx.items():
        for i, axis in enumerate(axes[2:]):
          if axis is not None:
            batch_idx[key] = (i, axis)
            break
        else:
          raise ValueError(f'Found no batch axis: {axes}.')

      # out axes
      if out_axes is None:
        out_axes = 0
      elif isinstance(out_axes, (int, dict)):
        out_axes = (out_axes, 0, 0)
      elif isinstance(out_axes, (tuple, list)):
        out_axes = tuple(out_axes) + (0, 0)
      assert isinstance(out_axes, (list, tuple))

      # reduce_func
      if reduce_func is None:
        reduce_func = lambda x: x.mean(axis=0)

      # jit function
      return _make_vmap(func=func,
                        nonbatched_vars=_dyn_vars,
                        batched_vars=_rand_vars,
                        in_axes=in_axes,
                        out_axes=out_axes,
                        axis_name=axis_name,
                        batch_idx=batch_idx)

  else:
    raise errors.BrainPyError(f'Only support instance of {Base.__name__}, or a callable '
                              f'function, but we got {type(func)}.')


def _device_reshape(x):
  """Reshape an input array in order to broadcast to multiple devices."""
  num_device = jax.local_device_count()

  if not hasattr(x, 'ndim'):
    raise errors.BrainPyError(f'Expected JaxArray, got {type(x)}. If you are trying to pass a scalar to '
                              f'parallel, first convert it to a JaxArray, for example np.float(0.5)')
  if x.ndim == 0:
    return np.broadcast_to(x, [num_device])
  if x.shape[0] % num_device != 0:
    raise errors.BrainPyError(f'Must be able to equally divide batch {x.shape} among '
                              f'{num_device} devices, but does not go equally.')
  return x.reshape((num_device, x.shape[0] // num_device) + x.shape[1:])


def _make_pmap(func, dyn_vars, rand_vars, reduce_func, axis_name=None, in_axes=0,
               out_axes=0, static_broadcasted_argnums=(), devices=None, backend=None,
               axis_size=None, donate_argnums=(), global_arg_shapes=None, f_name=None):
  @functools.partial(jax.pmap, in_axes=in_axes, out_axes=out_axes, axis_name=axis_name,
                     static_broadcasted_argnums=static_broadcasted_argnums, devices=devices,
                     backend=backend, axis_size=axis_size, donate_argnums=donate_argnums,
                     global_arg_shapes=global_arg_shapes)
  def pmapped_func(dyn_data, rand_data, *args, **kwargs):
    dyn_vars.assign(dyn_data)
    rand_vars.assign(rand_data)
    out = func(*args, **kwargs)
    dyn_changes = dyn_vars.dict()
    rand_changes = rand_vars.dict()
    return out, dyn_changes, rand_changes

  def call(*args):
    un_replicated = [k for k, v in dyn_vars.items()
                     if not isinstance(v.value, (ShardedDeviceArray, JaxprTracer, DynamicJaxprTracer))]
    if len(un_replicated):
      raise errors.BrainPyError(f'Some variables were not replicated: {un_replicated}.'
                                f'did you forget to call xx.replicate() on them?')
    _args = []
    for i, x in enumerate(args):
      if i + 2 in static_broadcasted_argnums:
        _args.append(x)
      else:
        _args.append(jax.tree_map(_device_reshape, [x])[0])
    dyn_data = dyn_vars.dict()
    rand_data = rand_vars.dict()
    output, dyn_changes, rand_changes = pmapped_func(dyn_data, rand_data, *_args)
    dyn_vars.assign(dyn_changes)
    rand_vars.assign(rand_changes)
    return jax.tree_map(reduce_func, output)

  return change_func_name(name=f_name, f=call) if f_name else call


def pmap(func, dyn_vars=None, axis_name=None, in_axes=0, out_axes=0, static_broadcasted_argnums=(),
         devices=None, backend=None, axis_size=None, donate_argnums=(), global_arg_shapes=None,
         reduce_func=None):
  """Parallel compilation for class objects.

  Parallel compile a function or a module to run on multiple devices in parallel.

  Parameters
  ----------
  func
  axis_name
  in_axes
  out_axes
  static_broadcasted_argnums
  devices
  backend
  axis_size
  donate_argnums
  global_arg_shapes

  Returns
  -------


  Examples
  --------


  """

  # if isinstance(func, DynamicalSystem):
  #   if len(func.steps):  # DynamicalSystem has step functions
  #
  #     # dynamical variables
  #     all_vars = (dyn_vars or func.vars().unique())
  #     dyn_vars = TensorCollector()
  #     rand_vars = TensorCollector()
  #     for key, val in all_vars.items():
  #       if isinstance(val, RandomState):
  #         rand_vars[key] = val
  #       else:
  #         dyn_vars[key] = val
  #
  #     # reduce function
  #     if reduce_func is None:
  #       reduce_func = jnp.concatenate
  #
  #     # static broadcast-ed arguments
  #     if static_broadcasted_argnums is None:
  #       static_broadcasted_argnums = ()
  #     elif isinstance(static_broadcasted_argnums, int):
  #       static_broadcasted_argnums = (static_broadcasted_argnums + 2,)
  #     elif isinstance(static_broadcasted_argnums, (tuple, list)):
  #       static_broadcasted_argnums = tuple(argnum + 2 for argnum in static_broadcasted_argnums)
  #     assert isinstance(static_broadcasted_argnums, (tuple, list))
  #
  #     # jit functions
  #     for key in func.steps.keys():
  #       step = func.steps[key]
  #       func.steps[key] = _make_pmap(dyn_vars=dyn_vars,
  #                                    rand_vars=rand_vars,
  #                                    func=step,
  #                                    axis_name=axis_name,
  #                                    in_axes=in_axes,
  #                                    out_axes=out_axes,
  #                                    static_broadcasted_argnums=static_broadcasted_argnums,
  #                                    devices=devices,
  #                                    backend=backend,
  #                                    axis_size=axis_size,
  #                                    donate_argnums=donate_argnums,
  #                                    global_arg_shapes=global_arg_shapes,
  #                                    reduce_func=reduce_func,
  #                                    f_name=key)
  #     return func

  if callable(func):
    if dyn_vars is not None:
      dyn_vars = dyn_vars
    elif isinstance(func, Base):  # Base has '__call__()' implementation
      dyn_vars = func.vars().unique()
    elif hasattr(func, '__self__'):
      if isinstance(func.__self__, Base):
        dyn_vars = func.__self__.vars().unique()

    if dyn_vars is None:
      return jax.pmap(func,
                      axis_name=axis_name,
                      in_axes=in_axes,
                      out_axes=out_axes,
                      static_broadcasted_argnums=static_broadcasted_argnums,
                      devices=devices,
                      backend=backend,
                      axis_size=axis_size,
                      donate_argnums=donate_argnums,
                      global_arg_shapes=global_arg_shapes)
    else:
      # dynamical variables
      dyn_vars = TensorCollector()
      rand_vars = TensorCollector()
      for key, val in dyn_vars.items():
        if isinstance(val, RandomState):
          rand_vars[key] = val
        else:
          dyn_vars[key] = val

      # static broadcast-ed arguments
      if static_broadcasted_argnums is None:
        static_broadcasted_argnums = ()
      elif isinstance(static_broadcasted_argnums, int):
        static_broadcasted_argnums = (static_broadcasted_argnums + 2,)
      elif isinstance(static_broadcasted_argnums, (tuple, list)):
        static_broadcasted_argnums = tuple(argnum + 2 for argnum in static_broadcasted_argnums)
      assert isinstance(static_broadcasted_argnums, (tuple, list))

      # reduce function
      if reduce_func is None:
        reduce_func = jnp.concatenate

      # jit function
      func.__call__ = _make_pmap(dyn_vars=dyn_vars,
                                 rand_vars=rand_vars,
                                 func=func,
                                 axis_name=axis_name,
                                 in_axes=in_axes,
                                 out_axes=out_axes,
                                 static_broadcasted_argnums=static_broadcasted_argnums,
                                 devices=devices,
                                 backend=backend,
                                 axis_size=axis_size,
                                 donate_argnums=donate_argnums,
                                 global_arg_shapes=global_arg_shapes,
                                 reduce_func=reduce_func)
      return func

  else:
    raise errors.BrainPyError(f'Only support instance of {Base.__name__}, or a callable function, '
                              f'but we got {type(func)}.')



# -*- coding: utf-8 -*-

from typing import Dict, Sequence, Any, Union

import jax.numpy as jnp

import brainpy.math as bm
from brainpy.errors import UnsupportedError
from brainpy.nn.base import Node, Network
from brainpy.types import Tensor
from brainpy.tools.checking import check_dict_data
from .rnn_runner import RNNRunner

__all__ = [
  'RNNTrainer',
]


class RNNTrainer(RNNRunner):
  """Structural Trainer for Models with Recurrent Dynamics."""

  train_nodes: Sequence[Node]  # need to be initialized by subclass
  train_pars: Dict[str, Any]  # need to be initialized by subclass

  def __init__(self, target, **kwargs):
    super(RNNTrainer, self).__init__(target=target, **kwargs)

    # get all trainable nodes
    self.train_nodes = self._get_trainable_nodes()

  def fit(
      self,
      train_data: Any,
      test_data: Any,
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
      initial_states: Union[Tensor, Dict[str, Tensor]] = None,
      initial_feedbacks: Dict[str, Tensor] = None,
      reset: bool = False,
      shared_kwargs: Dict = None
  ):  # need to be implemented by subclass
    raise NotImplementedError('Must implement the fit function. ')

  def _get_trainable_nodes(self):
    # check trainable nodes
    if isinstance(self.target, Network):
      train_nodes = [node for node in self.target.lnodes if node.trainable]
    elif isinstance(self.target, Node):
      train_nodes = [self.target] if self.target.trainable else []
    else:
      raise UnsupportedError('Must be a brainpy.nn.Node instance, '
                             f'while we got {type(self.target)}: {self.target}')
    return train_nodes

  def _check_ys(self, ys, num_batch, num_step, move_axis=False):
    # output_shapes = {}
    # for node in self.train_nodes:
    #   name = self.target.entry_nodes[0].name
    #   output_shapes[name] = node.output_shape

    if isinstance(ys, (bm.ndarray, jnp.ndarray)):
      if len(self.train_nodes) == 1:
        ys = {self.train_nodes[0].name: ys}
      else:
        raise ValueError(f'The network {self.target} has {len(self.train_nodes)} '
                         f'training nodes, while we only got one target data.')
    check_dict_data(ys, key_type=str, val_type=(bm.ndarray, jnp.ndarray))
    for key, val in ys.items():
      if val.ndim < 3:
        raise ValueError("Targets must be a tensor with shape of "
                         "(num_sample, num_time, feature_dim, ...), "
                         f"but we got {val.shape}")
      if val.shape[0] != num_batch:
        raise ValueError(f'Batch size of the target {key} does not match '
                         f'with the input data {val.shape[0]} != {num_batch}')
      if val.shape[1] != num_step:
        raise ValueError(f'The time step of the target {key} does not match '
                         f'with the input data {val.shape[1]} != {num_step})')
    if move_axis:
      # change shape to (num_time, num_sample, num_feature)
      ys = {k: bm.moveaxis(v, 0, 1) for k, v in ys.items()}
    return ys





# -*- coding: utf-8 -*-

from brainpy import math as bm, tools
from .base import InterLayerInitializer

__all__ = [
  'ZeroInit',
  'OneInit',
  'Identity',
]


class ZeroInit(InterLayerInitializer):
  """Zero initializer.

  Initialize the weights with zeros.
  """

  def __call__(self, shape, dtype=None):
    shape = [tools.size2num(d) for d in shape]
    return bm.zeros(shape, dtype=dtype)

  def __repr__(self):
    return self.__class__.__name__


class OneInit(InterLayerInitializer):
  """One initializer.

  Initialize the weights with the given values.

  Parameters
  ----------
  value : float, int, bm.ndarray
    The value to specify.
  """

  def __init__(self, value=1.):
    super(OneInit, self).__init__()
    self.value = value

  def __call__(self, shape, dtype=None):
    shape = [tools.size2num(d) for d in shape]
    return bm.ones(shape, dtype=dtype) * self.value

  def __repr__(self):
    return f'{self.__class__.__name__}(value={self.value})'


class Identity(InterLayerInitializer):
  """Returns the identity matrix.

  This initializer was proposed in (Le, et al., 2015) [1]_.

  Parameters
  ----------
  value : float
    The optional scaling factor.

  Returns
  -------
  shape: tuple of int
    The weight shape/size.

  References
  ----------
  .. [1] Le, Quoc V., Navdeep Jaitly, and Geoffrey E. Hinton. "A simple way to
         initialize recurrent networks of rectified linear units." arXiv preprint
         arXiv:1504.00941 (2015).
  """

  def __init__(self, value=1.):
    super(Identity, self).__init__()
    self.value = value

  def __call__(self, shape, dtype=None):
    if isinstance(shape, int):
      shape = (shape,)
    elif isinstance(shape, (tuple, list)):
      if len(shape) > 2:
        raise ValueError(f'Only support initialize 2D weights for {self.__class__.__name__}.')
    else:
      raise ValueError(f'Only support shape of int, or tuple/list of int '
                       f'in {self.__class__.__name__}, but we got {shape}.')
    shape = [tools.size2num(d) for d in shape]
    return bm.eye(*shape, dtype=dtype) * self.value

  def __repr__(self):
    return f'{self.__class__.__name__}(value={self.value})'



# -*- coding: utf-8 -*-


r"""This module provides commonly used activation functions.

Activation functions are a critical part of the design of a neural network.
The choice of activation function in the hidden layer will control how well
the network model learns the training dataset. The choice of activation
function in the output layer will define the type of predictions the model
can make.

"""

import operator

import jax
import jax.numpy as jnp
import jax.scipy
import numpy as np

from brainpy.math.jaxarray import JaxArray

__all__ = [
  'celu',
  'elu',
  'gelu',
  'glu',
  'hard_tanh',
  'hard_sigmoid',
  'hard_silu',
  'hard_swish',
  'leaky_relu',
  'log_sigmoid',
  'log_softmax',
  'one_hot',
  'normalize',
  'relu',
  'relu6',
  'sigmoid',
  'soft_sign',
  'softmax',
  'softplus',
  'silu',
  'swish',
  'selu',
]


def get(activation):
  global_vars = globals()

  if activation is None:
    return None

  if isinstance(activation, str):
    if activation not in global_vars:
      raise ValueError(f'Unknown activation function: {activation}, \nwe only support: '
                       f'{[k for k, v in global_vars.items() if not k.startswith("_") and callable(v)]}')
  elif callable(activation):
    return activation
  else:
    raise ValueError(f'Unknown activation function {activation}. ')
  return global_vars[activation]


tanh = jnp.tanh
identity = lambda x: x


def celu(x, alpha=1.0):
  r"""Continuously-differentiable exponential linear unit activation.

  Computes the element-wise function:

  .. math::
    \mathrm{celu}(x) = \begin{cases}
      x, & x > 0\\
      \alpha \left(\exp(\frac{x}{\alpha}) - 1\right), & x \le 0
    \end{cases}

  For more information, see
  `Continuously Differentiable Exponential Linear Units
  <https://arxiv.org/pdf/1704.07483.pdf>`_.

  Parameters
  ----------
  x : JaxArray, jnp.ndarray
    The input array.
  alpha : ndarray, float
    The default is 1.0.
  """
  x = x.value if isinstance(x, JaxArray) else x
  alpha = alpha.value if isinstance(alpha, JaxArray) else alpha
  return jnp.where(x > 0, x, alpha * jnp.expm1(x / alpha))


def elu(x, alpha=1.0):
  r"""Exponential linear unit activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{elu}(x) = \begin{cases}
      x, & x > 0\\
      \alpha \left(\exp(x) - 1\right), & x \le 0
    \end{cases}

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  alpha : scalar or JaxArray
    default: 1.0.
  """
  x = x.value if isinstance(x, JaxArray) else x
  alpha = alpha.value if isinstance(alpha, JaxArray) else alpha
  safe_x = jnp.where(x > 0, 0., x)
  return jnp.where(x > 0, x, alpha * jnp.expm1(safe_x))


def gelu(x, approximate=True):
  r"""Gaussian error linear unit activation function.

  If ``approximate=False``, computes the element-wise function:

  .. math::
    \mathrm{gelu}(x) = \frac{x}{2} \left(1 + \mathrm{erf} \left(
      \frac{x}{\sqrt{2}} \right) \right)

  If ``approximate=True``, uses the approximate formulation of GELU:

  .. math::
    \mathrm{gelu}(x) = \frac{x}{2} \left(1 + \mathrm{tanh} \left(
      \sqrt{\frac{2}{\pi}} \left(x + 0.044715 x^3 \right) \right) \right)

  For more information, see `Gaussian Error Linear Units (GELUs)
  <https://arxiv.org/abs/1606.08415>`_, section 2.

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  approximate: bool
    whether to use the approximate or exact formulation.
  """
  x = x.value if isinstance(x, JaxArray) else x
  if approximate:
    sqrt_2_over_pi = np.sqrt(2 / np.pi).astype(x.dtype)
    cdf = 0.5 * (1.0 + jnp.tanh(sqrt_2_over_pi * (x + 0.044715 * (x ** 3))))
    y = x * cdf
  else:
    y = jnp.array(x * (jax.lax.erf(x / np.sqrt(2)) + 1) / 2, dtype=x.dtype)
  return y


def glu(x, axis=-1):
  r"""Gated linear unit activation function.

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  axis: int
    The axis along which the split should be computed (default: -1)
  """
  size = x.shape[axis]
  assert size % 2 == 0, "axis size must be divisible by 2"
  x = x.value if isinstance(x, JaxArray) else x
  x1, x2 = jnp.split(x, 2, axis)
  return x1 * sigmoid(x2)


def hard_tanh(x):
  r"""Hard :math:`\mathrm{tanh}` activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{hard\_tanh}(x) = \begin{cases}
      -1, & x < -1\\
      x, & -1 \le x \le 1\\
      1, & 1 < x
    \end{cases}

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  x = x.value if isinstance(x, JaxArray) else x
  return jnp.where(x > 1, 1, jnp.where(x < -1, -1, x))


def hard_sigmoid(x):
  r"""Hard Sigmoid activation function.

  Computes the element-wise function

  .. math::
    \mathrm{hard\_sigmoid}(x) = \frac{\mathrm{relu6}(x + 3)}{6}

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  return relu6(x + 3.) / 6.


def hard_silu(x):
  r"""Hard SiLU activation function

  Computes the element-wise function

  .. math::
    \mathrm{hard\_silu}(x) = x \cdot \mathrm{hard\_sigmoid}(x)

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  return x * hard_sigmoid(x)


hard_swish = hard_silu


def leaky_relu(x, negative_slope=1e-2):
  r"""Leaky rectified linear unit activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{leaky\_relu}(x) = \begin{cases}
      x, & x \ge 0\\
      \alpha x, & x < 0
    \end{cases}

  where :math:`\alpha` = :code:`negative_slope`.

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  negative_slope : float
    The scalar specifying the negative slope (default: 0.01)
  """
  x = x.value if isinstance(x, JaxArray) else x
  return jnp.where(x >= 0, x, negative_slope * x)


def softplus(x):
  r"""Softplus activation function.

  Computes the element-wise function

  .. math::
    \mathrm{softplus}(x) = \log(1 + e^x)

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  x = x.value if isinstance(x, JaxArray) else x
  return jnp.logaddexp(x, 0)


def log_sigmoid(x):
  r"""Log-sigmoid activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{log\_sigmoid}(x) = \log(\mathrm{sigmoid}(x)) = -\log(1 + e^{-x})

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  return -softplus(-x)


def log_softmax(x, axis=-1):
  r"""Log-Softmax function.

  Computes the logarithm of the :code:`softmax` function, which rescales
  elements to the range :math:`[-\infty, 0)`.

  .. math ::
    \mathrm{log\_softmax}(x) = \log \left( \frac{\exp(x_i)}{\sum_j \exp(x_j)}
    \right)

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  axis: int
    The axis or axes along which the :code:`log_softmax` should be
    computed. Either an integer or a tuple of integers.
  """
  x = x.value if isinstance(x, JaxArray) else x
  shifted = x - jax.lax.stop_gradient(x.max(axis, keepdims=True))
  return shifted - jnp.log(jnp.sum(jnp.exp(shifted), axis, keepdims=True))


def _canonicalize_axis(axis, num_dims) -> int:
  """Canonicalize an axis in [-num_dims, num_dims) to [0, num_dims)."""
  axis = operator.index(axis)
  if not -num_dims <= axis < num_dims:
    raise ValueError(
      "axis {} is out of bounds for array of dimension {}".format(
        axis, num_dims))
  if axis < 0:
    axis = axis + num_dims
  return axis


def one_hot(x, num_classes, *, dtype=None, axis=-1):
  r"""One-hot encodes the given indicies.

  Each index in the input ``x`` is encoded as a vector of zeros of length
  ``num_classes`` with the element at ``index`` set to one::

    >>> import jax.numpy as jnp
    >>> one_hot(jnp.array([0, 1, 2]), 3)
    DeviceArray([[1., 0., 0.],
                  [0., 1., 0.],
                  [0., 0., 1.]], dtype=float32)

  Indicies outside the range [0, num_classes) will be encoded as zeros::

    >>> import jax.numpy as jnp
    >>> one_hot(jnp.array([-1, 3]), 3)
    DeviceArray([[0., 0., 0.],
                 [0., 0., 0.]], dtype=float32)

  Args:
    x: A tensor of indices.
    num_classes: Number of classes in the one-hot dimension.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    axis: the axis or axes along which the function should be
      computed.
  """
  num_classes = jax.core.concrete_or_error(
    int, num_classes, "The error arose in jax.nn.one_hot argument `num_classes`.")
  dtype = jax.dtypes.canonicalize_dtype(jnp.float64 if dtype is None else dtype)
  x = jnp.asarray(x)
  try:
    output_pos_axis = _canonicalize_axis(axis, x.ndim + 1)
  except TypeError:
    axis_size = jax.lax.psum(1, axis)
    if num_classes != axis_size:
      raise ValueError(f"Expected num_classes to match the size of axis {axis}, "
                       f"but {num_classes} != {axis_size}") from None
    axis_idx = jax.lax.axis_index(axis)
    return jnp.asarray(x == axis_idx, dtype=dtype)
  axis = operator.index(axis)
  lhs = jax.lax.expand_dims(x, (axis,))
  rhs_shape = [1] * x.ndim
  rhs_shape.insert(output_pos_axis, num_classes)
  rhs = jax.lax.broadcast_in_dim(jnp.arange(num_classes, dtype=x.dtype),
                                 rhs_shape, (output_pos_axis,))
  return jnp.asarray(lhs == rhs, dtype=dtype)


def normalize(x, axis=-1, mean=None, variance=None, epsilon=1e-5):
  """Normalizes an array by subtracting mean and dividing by sqrt(var)."""
  x = x.value if isinstance(x, JaxArray) else x
  if mean is None:
    mean = jnp.mean(x, axis, keepdims=True)
  if variance is None:
    # this definition is traditionally seen as less accurate than jnp.var's
    # mean((x - mean(x))**2) but may be faster and even, given typical
    # activation distributions and low-precision arithmetic, more accurate
    # when used in neural network normalization layers
    variance = jnp.mean(jnp.square(x), axis, keepdims=True) - jnp.square(mean)
  y = (x - mean) * jax.lax.rsqrt(variance + epsilon)
  return y


def relu(x):
  x = x.value if isinstance(x, JaxArray) else x
  return jax.nn.relu(x)


def relu6(x):
  r"""Rectified Linear Unit 6 activation function.

  Computes the element-wise function

  .. math::
    \mathrm{relu6}(x) = \min(\max(x, 0), 6)

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  x = x.value if isinstance(x, JaxArray) else x
  return jnp.minimum(jnp.maximum(x, 0), 6.)


def sigmoid(x):
  r"""Sigmoid activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{sigmoid}(x) = \frac{1}{1 + e^{-x}}

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  x = x.value if isinstance(x, JaxArray) else x
  return jax.scipy.special.expit(x)


def soft_sign(x):
  r"""Soft-sign activation function.

  Computes the element-wise function

  .. math::
    \mathrm{soft\_sign}(x) = \frac{x}{|x| + 1}

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  x = x.value if isinstance(x, JaxArray) else x
  return x / (jnp.abs(x) + 1)


def softmax(x, axis=-1):
  r"""Softmax function.

  Computes the function which rescales elements to the range :math:`[0, 1]`
  such that the elements along :code:`axis` sum to :math:`1`.

  .. math ::
    \mathrm{softmax}(x) = \frac{\exp(x_i)}{\sum_j \exp(x_j)}

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  axis: int
    The axis or axes along which the softmax should be computed. The
    softmax output summed across these dimensions should sum to :math:`1`.
    Either an integer or a tuple of integers.
  """
  x = x.value if isinstance(x, JaxArray) else x
  unnormalized = jnp.exp(x - jax.lax.stop_gradient(x.max(axis, keepdims=True)))
  return unnormalized / unnormalized.sum(axis, keepdims=True)


def silu(x):
  r"""SiLU activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{silu}(x) = x \cdot \mathrm{sigmoid}(x) = \frac{x}{1 + e^{-x}}

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  x = x.value if isinstance(x, JaxArray) else x
  return x * sigmoid(x)


swish = silu


def selu(x):
  r"""Scaled exponential linear unit activation.

  Computes the element-wise function:

  .. math::
    \mathrm{selu}(x) = \lambda \begin{cases}
      x, & x > 0\\
      \alpha e^x - \alpha, & x \le 0
    \end{cases}

  where :math:`\lambda = 1.0507009873554804934193349852946` and
  :math:`\alpha = 1.6732632423543772848170429916717`.

  For more information, see
  `Self-Normalizing Neural Networks
  <https://papers.nips.cc/paper/6698-self-normalizing-neural-networks.pdf>`_.

  Parameters
  ----------
  x: JaxArray, jnp.ndarray
    The input array.
  """
  alpha = 1.6732632423543772848170429916717
  scale = 1.0507009873554804934193349852946
  x = x.value if isinstance(x, JaxArray) else x
  safe_x = jnp.where(x > 0, 0., x)
  return scale * jnp.where(x > 0, x, alpha * jnp.expm1(safe_x))



# -*- coding: utf-8 -*-


from collections.abc import Iterable
from typing import Union, Callable
import jax.numpy as jnp
import numpy as np

from brainpy import math as bm
from brainpy.dyn.base import DynamicalSystem
from brainpy.errors import RunningError
from brainpy.running.monitor import Monitor
from brainpy.initialize import init_param, Initializer

__all__ = [
  'size2len',
  'init_delay',
  'check_and_format_inputs',
  'check_and_format_monitors',
]

SUPPORTED_INPUT_OPS = ['-', '+', '*', '/', '=']
SUPPORTED_INPUT_TYPE = ['fix', 'iter', 'func']


def init_delay(delay_step: Union[int, bm.ndarray, jnp.ndarray, Callable, Initializer],
               delay_target: Union[bm.ndarray, jnp.ndarray],
               delay_data: Union[bm.ndarray, jnp.ndarray] = None):
  """Initialize delay variable.

  Parameters
  ----------
  delay_step: int, ndarray, JaxArray
    The number of delay steps. It can an integer of an array of integers.
  delay_target: ndarray, JaxArray
    The target variable to delay.
  delay_data: optional, ndarray, JaxArray
    The initial delay data.

  Returns
  -------
  info: tuple
    The triple of delay type, delay steps, and delay variable.
  """
  # check delay type
  if delay_step is None:
    delay_type = 'none'
  elif isinstance(delay_step, int):
    delay_type = 'homo'
  elif isinstance(delay_step, (bm.ndarray, jnp.ndarray, np.ndarray)):
    delay_type = 'heter'
    delay_step = bm.asarray(delay_step)
  elif callable(delay_step):
    delay_step = init_param(delay_step, delay_target.shape, allow_none=False)
    delay_type = 'heter'
  else:
    raise ValueError(f'Unknown "delay_steps" type {type(delay_step)}, only support '
                     f'integer, array of integers, callable function, brainpy.init.Initializer.')
  if delay_type == 'heter':
    if delay_step.dtype not in [bm.int32, bm.int64]:
      raise ValueError('Only support delay steps of int32, int64. If your '
                       'provide delay time length, please divide the "dt" '
                       'then provide us the number of delay steps.')
    if delay_target.shape[0] != delay_step.shape[0]:
      raise ValueError(f'Shape is mismatched: {delay_target.shape[0]} != {delay_step.shape[0]}')

  # init delay data
  if delay_type == 'homo':
    delays = bm.LengthDelay(delay_target, delay_step, initial_delay_data=delay_data)
  elif delay_type == 'heter':
    if delay_step.size != delay_target.size:
      raise ValueError('Heterogeneous delay must have a length '
                       f'of the delay target {delay_target.shape}, '
                       f'while we got {delay_step.shape}')
    delays = bm.LengthDelay(delay_target, int(delay_step.max()))
  else:
    delays = None

  return delay_type, delay_step, delays


def size2len(size):
  if isinstance(size, int):
    return size
  elif isinstance(size, (tuple, list)):
    a = 1
    for b in size:
      a *= b
    return a
  else:
    raise ValueError


def check_and_format_inputs(host, inputs):
  """Check inputs and get the formatted inputs for the given population.

  Parameters
  ----------
  host : DynamicalSystem
      The host which contains all data.
  inputs : tuple, list
      The inputs of the population.

  Returns
  -------
  formatted_inputs : tuple, list
      The formatted inputs of the population.
  """

  # 1. check inputs
  # ---------
  if inputs is None:
    inputs = []
  if not isinstance(inputs, (tuple, list)):
    raise RunningError('"inputs" must be a tuple/list.')
  if len(inputs) > 0 and not isinstance(inputs[0], (list, tuple)):
    if isinstance(inputs[0], str):
      inputs = [inputs]
    else:
      raise RunningError('Unknown input structure, only support inputs '
                         'with format of "(target, value, [type, operation])".')
  for one_input in inputs:
    if not 2 <= len(one_input) <= 4:
      raise RunningError('For each target, you must specify '
                         '"(target, value, [type, operation])".')
    if len(one_input) == 3 and one_input[2] not in SUPPORTED_INPUT_TYPE:
      raise RunningError(f'Input type only supports '
                         f'"{SUPPORTED_INPUT_TYPE}", '
                         f'not "{one_input[2]}".')
    if len(one_input) == 4 and one_input[3] not in SUPPORTED_INPUT_OPS:
      raise RunningError(f'Input operation only supports '
                         f'"{SUPPORTED_INPUT_OPS}", '
                         f'not "{one_input[3]}".')

  # 2. get targets and attributes
  # ---------
  inputs_which_found_target = []
  inputs_not_found_target = []

  # checking 1: absolute access
  #    Check whether the input target node is accessible,
  #    and check whether the target node has the attribute
  nodes = host.nodes(method='absolute')
  nodes[host.name] = host
  for one_input in inputs:
    key = one_input[0]
    if not isinstance(key, str):
      raise RunningError(f'For each input, input[0] must be a string  to '
                         f'specify variable of the target, but we got {key}.')
    splits = key.split('.')
    target = '.'.join(splits[:-1])
    key = splits[-1]
    if target == '':
      real_target = host
    else:
      if target not in nodes:
        inputs_not_found_target.append(one_input)
        continue
      real_target = nodes[target]
    if not hasattr(real_target, key):
      raise RunningError(f'Input target key "{key}" is not defined in {real_target}.')
    inputs_which_found_target.append((real_target, key) + tuple(one_input[1:]))

  # checking 2: relative access
  #    Check whether the input target node is accessible
  #    and check whether the target node has the attribute
  if len(inputs_not_found_target):
    nodes = host.nodes(method='relative')
    for one_input in inputs_not_found_target:
      splits = one_input[0].split('.')
      target, key = '.'.join(splits[:-1]), splits[-1]
      if target not in nodes:
        raise RunningError(f'Input target "{target}" is not defined in {host}.')
      real_target = nodes[target]
      if not hasattr(real_target, key):
        raise RunningError(f'Input target key "{key}" is not defined in {real_target}.')
      inputs_which_found_target.append((real_target, key) + tuple(one_input[1:]))

  # 3. format inputs
  # ---------
  formatted_inputs = []
  for one_input in inputs_which_found_target:
    # input value
    data_value = one_input[2]

    # input type
    if len(one_input) >= 4:
      if one_input[3] == 'iter':
        if not isinstance(data_value, Iterable):
          raise ValueError(f'Input "{data_value}" for "{one_input[0]}.{one_input[1]}" '
                           f'is set to be "iter" type, however we got the value with '
                           f'the type of {type(data_value)}')
      elif one_input[3] == 'func':
        if not callable(data_value):
          raise ValueError(f'Input "{data_value}" for "{one_input[0]}.{one_input[1]}" '
                           f'is set to be "func" type, however we got the value with '
                           f'the type of {type(data_value)}')
      elif one_input[3] != 'fix':
        raise RunningError(f'Only support {SUPPORTED_INPUT_TYPE} input type, but '
                           f'we got "{one_input[3]}" in {one_input}')

      data_type = one_input[3]
    else:
      data_type = 'fix'

    # operation
    if len(one_input) == 5:
      data_op = one_input[4]
    else:
      data_op = '+'
    if data_op not in SUPPORTED_INPUT_OPS:
      raise RunningError(f'Only support {SUPPORTED_INPUT_OPS}, while we got '
                         f'{data_op} in {one_input}')

    # final
    format_inp = one_input[:2] + (data_value, data_type, data_op)
    formatted_inputs.append(format_inp)

  return formatted_inputs


def check_and_format_monitors(host, mon):
  """Return a formatted monitor items:

  >>> [(node, key, target, variable, idx, interval),
  >>>  ...... ]

  """
  assert isinstance(host, DynamicalSystem)
  assert isinstance(mon, Monitor)

  formatted_mon_items = []

  # master node:
  #    Check whether the input target node is accessible,
  #    and check whether the target node has the attribute
  name2node = {node.name: node for node in list(host.nodes().unique().values())}
  for key, idx, interval in zip(mon.item_names, mon.item_indices, mon.item_intervals):
    # target and variable
    splits = key.split('.')
    if len(splits) == 1:
      if not hasattr(host, splits[0]):
        raise RunningError(f'{host} does not has variable {key}.')
      target = host
      variable = splits[-1]
    else:
      if not hasattr(host, splits[0]):
        if splits[0] not in name2node:
          raise RunningError(f'Cannot find target {key} in monitor of {host}, please check.')
        else:
          target = name2node[splits[0]]
          assert len(splits) == 2
          variable = splits[-1]
      else:
        target = host
        for s in splits[:-1]:
          try:
            target = getattr(target, s)
          except KeyError:
            raise RunningError(f'Cannot find {key} in {host}, please check.')
        variable = splits[-1]

    # idx
    if isinstance(idx, int): idx = bm.array([idx])

    # interval
    if interval is not None:
      if not isinstance(interval, float):
        raise RunningError(f'"interval" must be a float (denotes time), but we got {interval}')

    # append
    formatted_mon_items.append((key, target, variable, idx, interval,))

  return formatted_mon_items



# -*- coding: utf-8 -*-

import warnings
from typing import Union, Sequence, Dict, Any, Callable, Optional

import jax.numpy as jnp

import brainpy.math as bm
from brainpy.initialize import Initializer, init_param as true_init_param
from brainpy.tools.checking import check_dict_data
from brainpy.types import Tensor, Shape

__all__ = [
  'tensor_sum',
  'init_param',
  'check_data_batch_size',
  'check_rnn_data_time_step',
  'serialize_kwargs',
]


def tensor_sum(values: Union[Sequence[Tensor], Dict[Any, Tensor], Tensor]):
  if isinstance(values, (bm.ndarray, jnp.ndarray)):
    return values
  if isinstance(values, dict):
    values = list(values.values())
  elif isinstance(values, (tuple, list)):
    values = list(values)
  else:
    raise ValueError('Unknown types of tensors.')
  res = values[0]
  for v in values[1:]:
    res = res + v
  return res


def init_param(param: Union[Callable, Initializer, bm.ndarray, jnp.ndarray],
               size: Shape):
  """Initialize parameters.

  .. deprecated:: 2.1.2
     Please use "brainpy.init.init_param" instead.

  Parameters
  ----------
  param: callable, Initializer, bm.ndarray, jnp.ndarray
    The initialization of the parameter.
    - If it is None, the created parameter will be None.
    - If it is a callable function :math:`f`, the ``f(size)`` will be returned.
    - If it is an instance of :py:class:`brainpy.init.Initializer``, the ``f(size)`` will be returned.
    - If it is a tensor, then this function check whether ``tensor.shape`` is equal to the given ``size``.
  size: int, sequence of int
    The shape of the parameter.
  """
  warnings.warn('Please use "brainpy.init.init_param" instead. '
                '"brainpy.nn.init_param" is deprecated since version 2.1.2. ',
                DeprecationWarning)
  return true_init_param(param, size)


def check_data_batch_size(data: Dict, num_batch=None):
  if len(data) == 1:
    batch_size = list(data.values())[0].shape[0]
  else:
    batches = []
    for key, val in data.items():
      batches.append(val.shape[0])
    if len(set(batches)) != 1:
      raise ValueError('Batch sizes are not consistent among the given data. '
                       f'Got {set(batches)}. We expect only one batch size.')
    batch_size = batches[0]
  if (num_batch is not None) and batch_size != num_batch:
    raise ValueError(f'Batch size is not consistent with the expected {batch_size} != {num_batch}')
  return batch_size


def check_rnn_data_time_step(data: Dict, num_step=None):
  if len(data) == 1:
    time_step = list(data.values())[0].shape[1]
  else:
    steps = []
    for key, val in data.items():
      steps.append(val.shape[1])
    if len(set(steps)) != 1:
      raise ValueError('Time steps are not consistent among the given data. '
                       f'Got {set(steps)}. We expect only one time step.')
    time_step = steps[0]
  if (num_step is not None) and time_step != num_step:
    raise ValueError(f'Time step is not consistent with the expected {time_step} != {num_step}')
  return time_step


def serialize_kwargs(shared_kwargs: Optional[Dict]):
  """Serialize kwargs."""
  shared_kwargs = dict() if shared_kwargs is None else shared_kwargs
  check_dict_data(shared_kwargs,
                  key_type=str,
                  val_type=(bool, float, int, complex),
                  name='shared_kwargs')
  shared_kwargs = {key: shared_kwargs[key] for key in sorted(shared_kwargs.keys())}
  return str(shared_kwargs)



# -*- coding: utf-8 -*-

from brainpy import math
from brainpy.integrators import constants, utils

__all__ = [
  'srk1_strong',
]


_SDE_UNKNOWN_NO = 0


def basic_info(f, g):
  vdt = 'dt'
  if f.__name__.isidentifier():
    func_name = f.__name__
  elif g.__name__.isidentifier():
    func_name = g.__name__
  else:
    global _SDE_UNKNOWN_NO
    func_name = f'unknown_sde{_SDE_UNKNOWN_NO}'
  func_new_name = constants.SDE_INT + func_name
  variables, parameters, arguments = utils.get_args(f)
  return vdt, variables, parameters, arguments, func_new_name


def _vector_wiener_terms(code_lines, sde_type, vdt, shape_D, shape_m):
  if sde_type == constants.ITO_SDE:
    I2 = f'0.5*(_term3 - {vdt} * math.eye({shape_m})) + _a*0.5*{vdt}/math.pi'
  elif sde_type == constants.STRA_SDE:
    I2 = f'0.5*_term3 + _a*0.5*dt/math.pi'
  else:
    raise ValueError(f'Unknown SDE_INT type: {sde_type}. We only supports {constants.SUPPORTED_INTG_TYPE}.')

  if shape_D:
    shape_D = shape_D + '+'

  noise_string = f'''
  # Noise Terms #
  # ----------- #
    
  # single Ito integrals
  _I1 = math.normal(0., {vdt}_sqrt, {shape_D}({shape_m},))
  # double Ito integrals
  _h = (2.0 / {vdt}) ** 0.5)
  _a = math.zeros(shape={shape_D}({shape_m}, {shape_m}))
  for _k in range(1, num_iter + 1):
    _x = math.normal(loc=0., scale=1., size={shape_D}({shape_m}, 1))
    _y = math.normal(loc=0., scale=1., size={shape_D}(1, {shape_m})) + _h * _I1
    _term1 = math.matmul(_x, _y)
    _term2 = math.matmul(math.reshape(_y, {shape_D}({shape_m}, 1)), 
                         math.reshape(_x, {shape_D}(1, {shape_m})))
    _a += (_term1 - _term2) / _k
  _I1_rs = math.reshape(_I1, {shape_D}({shape_m}, 1))
  _term3 = math.matmul(_I1_rs, math.reshape(_I1, {shape_D}(1, {shape_m})))
  _I2 = {I2}
  '''
  noise_lines = noise_string.split('\n')
  code_lines.extend(noise_lines)


# ----------
# Wrapper
# ----------


def _srk2_pop_var_vector_wiener(sde_type, code_lines, variables, parameters, vdt):
  # shape information
  # -----
  all_f = [f'f_{var}' for var in variables]
  all_g = [f'g_{var}' for var in variables]
  noise_string = f'''
  {", ".join(all_f)} = f({", ".join(variables + parameters)})  # shape = (..)
  {", ".join(all_g)} = g({", ".join(variables + parameters)})  # shape = (.., m)
  noise_shape = math.shape(g_x1)
  _D = noise_shape[:-1]
  _m = noise_shape[-1]
  '''
  code_lines.extend(noise_string.split("\n"))

  # noise terms
  _vector_wiener_terms(code_lines, sde_type, vdt, shape_D='_D', shape_m='_m')

  # numerical integration
  # step 1
  # ---
  # g_x1_rs = math.reshape(g_x1, _D + (1, _m))
  # g_x2_rs = math.reshape(g_x2, _D + (1, _m))
  for var in variables:
    code_lines.append(f"  g_{var}_rs = math.reshape(g_{var}, _D+(1, _m))")
  # step 2
  # ---
  # g_H1_x1 = math.reshape(math.matmul(g_x1_rs, _I2) / dt_sqrt, _D + (_m,))
  # g_H1_x2 = math.reshape(math.matmul(g_x2_rs, _I2) / dt_sqrt, _D + (_m,))
  for var in variables:
    code_lines.append(f'  g_H1_{var} = math.reshape(math.matmul(g_{var}_rs, _I2) / {vdt}_sqrt, _D + (_m,))')
  # step 3
  # ---
  # x1_rs = math.reshape(x1, _D + (1,))
  # x2_rs = math.reshape(x2, _D + (1,))
  for var in variables:
    code_lines.append(f'  {var}_rs = math.reshape({var}, _D + (1,))')
  # step 4
  # ---
  # H2_x1 = x1_rs + g_H1_x1
  # H3_x1 = x1_rs - g_H1_x1
  for var in variables:
    code_lines.append(f'  H2_{var} = {var}_rs + g_H1_{var}')
    code_lines.append(f'  H3_{var} = {var}_rs - g_H1_{var}')
  code_lines.append('  ')
  # step 5
  # ---
  # _g_x1 = math.matmul(g_x1_rs, _I1_rs)
  for var in variables:
    code_lines.append(f'  _g_{var} = math.matmul(g_{var}_rs, _I1_rs)')
  # step 6
  # ----
  # x1_new = x1 + f_x1 + _g_x1[..., 0, 0]
  for var in variables:
    code_lines.append(f'  {var}_new = {var} + f_{var} + _g_{var}[..., 0, 0]')
  # for _k in range(_m):
  code_lines.append('for _k in range(_m):')
  #   g_x1_H2, g_x2_H2 = g(H2_x1[..., _k], H2_x2[..., _k], t, *args)
  all_H2 = [f'H2_{var}[..., _k]' for var in variables]
  all_g_H2 = [f'g_{var}_H2' for var in variables]
  code_lines.append(f'    {", ".join(all_g_H2)} = g({", ".join(all_H2 + parameters)})')
  #   g_x1_H3, g_x2_H3 = g(H3_x1[..., _k], H3_x2[..., _k], t, *args)
  all_H3 = [f'H3_{var}[..., _k]' for var in variables]
  all_g_H3 = [f'g_{var}_H3' for var in variables]
  code_lines.append(f'    {", ".join(all_g_H3)} = g({", ".join(all_H3 + parameters)})')
  #   x1_new += 0.5 * dt_sqrt * (g_x1_H2[..., _k] - g_x1_H3[..., _k])
  #   x2_new += 0.5 * dt_sqrt * (g_x2_H2[..., _k] - g_x2_H3[..., _k])
  for var in variables:
    code_lines.append(f'    {var}_new += 0.5 * {vdt}_sqrt * (g_{var}_H2[..., _k] - g_{var}_H3[..., _k])')


def _srk2_pop_or_scalar_var_scalar_wiener(sde_type, code_lines, variables, parameters, vdt):
  if sde_type == constants.ITO_SDE:
    I2 = f'0.5 * (_I1 * _I1 - {vdt})'
  elif sde_type == constants.STRA_SDE:
    I2 = f'0.5 * _I1 * _I1'
  else:
    raise ValueError(f'Unknown SDE_INT type: {sde_type}. We only supports {constants.SUPPORTED_INTG_TYPE}.')

  # shape info
  # -----
  all_f = [f'f_{var}' for var in variables]
  all_g = [f'g_{var}' for var in variables]

  code_string = f'''
  {", ".join(all_f)} = f({", ".join(variables + parameters)})  # shape = (..)
  {", ".join(all_g)} = g({", ".join(variables + parameters)})  # shape = (..)

  # single Ito integrals
  _I1 = math.normal(0., {vdt}_sqrt, math.shape({variables[0]}))  # shape = (..)
  # double Ito integrals
  _I2 = {I2}  # shape = (..)
  '''
  code_splits = code_string.split('\n')
  code_lines.extend(code_splits)

  # numerical integration
  # -----
  # H1
  for var in variables:
    code_lines.append(f'  g_H1_{var} = g_{var} * _I2 / {vdt}_sqrt  # shape (.., )')
  # H2
  all_H2 = [f'H2_{var}' for var in variables]
  for var in variables:
    code_lines.append(f'  H2_{var} = {var} + g_H1_{var}  # shape (.., )')
  all_g_H2 = [f'g_{var}_H2' for var in variables]
  code_lines.append(f'  {", ".join(all_g_H2)} = g({", ".join(all_H2 + parameters)})')
  code_lines.append(f'  ')
  # H3
  all_H3 = [f'H3_{var}' for var in variables]
  for var in variables:
    code_lines.append(f'  H3_{var} = {var} - g_H1_{var}  # shape (.., )')
  all_g_H3 = [f'g_{var}_H3' for var in variables]
  code_lines.append(f'  {", ".join(all_g_H3)} = g({", ".join(all_H3 + parameters)})')
  code_lines.append(f'  ')
  # final results
  for var in variables:
    code_lines.append(f'  {var}_new = {var} + f_{var} + g_{var} * _I1 '
                      f'+ 0.5 * {vdt}_sqrt * (g_{var}_H2 - g_{var}_H3)')


def _srk1_scalar_var_with_vector_wiener(sde_type, code_lines, variables, parameters, vdt):
  # shape information
  all_f = [f'f_{var}' for var in variables]
  all_g = [f'g_{var}' for var in variables]
  code1 = f'''
  # shape info #
  # ---------- #

  {", ".join(all_f)} = f({", ".join(variables + parameters)})  # shape = ()
  {", ".join(all_g)} = g({", ".join(variables + parameters)})  # shape = (m)
  noise_shape = math.shape(g_x1)
  _m = noise_shape[0]
  '''
  code_lines.extend(code1.split('\n'))

  # noise term
  _vector_wiener_terms(code_lines, sde_type, vdt, shape_D='', shape_m='_m')

  # numerical integration

  # p1
  # ---
  # g_x1_rs = math.reshape(g_x1, (1, _m))
  # g_x2_rs = math.reshape(g_x2, (1, _m))
  for var in variables:
    code_lines.append(f'  g_{var}_rs = math.reshape(g_{var}, (1, _m))')

  # p2
  # ---
  # g_H1_x1 = math.matmul(g_x1_rs, _I2) / dt_sqrt  # shape (1, m)
  # g_H1_x2 = math.matmul(g_x2_rs, _I2) / dt_sqrt  # shape (1, m)
  for var in variables:
    code_lines.append(f'  g_H1_{var} = math.matmul(g_{var}_rs, _I2) / {vdt}_sqrt  # shape (1, m)')

  # p3
  # ---
  # H2_x1 = x1 + g_H1_x1[0]  # shape (m)
  # H3_x1 = x1 - g_H1_x1[0]  # shape (m)
  for var in variables:
    code_lines.append(f'  H2_{var} = {var} + g_H1_{var}[0]  # shape (m)')
  code_lines.append('  ')

  # p4
  # ---
  # g1_x1 = math.matmul(g_x1_rs, _I1_rs)  # shape (1, 1)
  # x1_new = x1 + f_x1 + g1_x1[0, 0]  # shape ()
  for var in variables:
    code_lines.append(f'  g1_{var} = math.matmul(g_{var}_rs, _I1_rs)  # shape (1, 1)')
    code_lines.append(f'  {var}_new = {var} + f_{var} + g1_{var}[0, 0]  # shape ()')

  # p5
  # ---
  # for _k in range(_m):
  #    g_x1_H2, g_x2_H2 = g(H2_x1[_k], H2_x2[_k], t, *args)
  #    g_x1_H3, g_x2_H3 = g(H3_x1[_k], H3_x2[_k], t, *args)
  #    x1_new += 0.5 * dt_sqrt * (g_x1_H2[_k] - g_x1_H3[_k])
  #    x2_new += 0.5 * dt_sqrt * (g_x2_H2[_k] - g_x2_H3[_k])
  code_lines.append('  for _k in range(_m):')
  all_h2_k = [f'H2_{var}[_k]' for var in variables]
  all_g_h2 = [f'g_{var}_H2' for var in variables]
  code_lines.append(f'    {", ".join(all_g_h2)} = g({", ".join(all_h2_k + parameters)})')
  all_h3_k = [f'H3_{var}[_k]' for var in variables]
  all_g_h3 = [f'g_{var}_H3' for var in variables]
  code_lines.append(f'    {", ".join(all_g_h3)} = g({", ".join(all_h3_k + parameters)})')
  for var in variables:
    code_lines.append(f'    {var}_new += 0.5 * {vdt}_sqrt * (g_{var}_H2[_k] - g_{var}_H3[_k])')


def _srk1_system_var_with_vector_wiener(sde_type, code_lines, variables, parameters, vdt):
  # shape information
  code1 = f'''
  # shape infor #
  # ----------- #
    
  f_x = f({", ".join(variables + parameters)})  # shape = (d, ..)
  g_x = g({", ".join(variables + parameters)})  # shape = (d, .., m)
  _shape = math.shape(g_x)
  _d = _shape[0]
  _m = _shape[-1]
  _D = _shape[1:-1]
  '''
  code_lines.extend(code1.split('\n'))

  # noise term
  _vector_wiener_terms(code_lines, sde_type, vdt, shape_D='_D', shape_m='_m')

  # numerical integration
  code2 = f'''
  # numerical integration #
  # --------------------- #
  
  g_x2 = math.moveaxis(g_x, 0, -2)  # shape = (.., d, m)
  g_H1_k = math.matmul(g_x2, _I2) / dt_sqrt  # shape (.., d, m)
  g_H1_k = math.moveaxis(g_H1_k, -2, 0)  # shape (d, .., m)
  x_rs = math.reshape(x, (_d,) + _D + (1,))
  H2 = x_rs + g_H1_k  # shape (d, .., m)
  H3 = x_rs - g_H1_k  # shape (d, .., m)
  
  g1 = math.matmul(g_x2, _I1_rs)  # shape (.., d, 1)
  g1 = math.moveaxis(g1, -2, 0)  # shape (d, .., 1)
  y = x + f_x + g1[..., 0]  # shape (d, ..)
  for _k in range(_m):
    y += 0.5 * dt_sqrt * g(H2[..., _k], t, *args)[..., _k]
    y -= 0.5 * dt_sqrt * g(H3[..., _k], t, *args)[..., _k]
  '''
  code_lines.extend(code2.split('\n'))


def _srk1_system_var_with_scalar_wiener(sde_type, code_lines, variables, parameters, vdt):
  if sde_type == constants.ITO_SDE:
    I2 = f'0.5 * (_I1 * _I1 - {vdt})'
  elif sde_type == constants.STRA_SDE:
    I2 = f'0.5 * _I1 * _I1'
  else:
    raise ValueError(f'Unknown SDE_INT type: {sde_type}. We only supports {constants.SUPPORTED_INTG_TYPE}.')

  code_string = f'''
  f_x = f({", ".join(variables + parameters)})  # shape = (d, ..)
  g_x = g({", ".join(variables + parameters)})  # shape = (d, ..)
  _shape = math.shape(g_x)
  _d = _shape[0]
  _D = _shape[1:]

  # single Ito integrals
  _I1 = math.normal(0., {vdt}_sqrt, _D)  # shape = (..)
  # double Ito integrals
  _I2 = {I2}  # shape = (..)

  # numerical integration #
  # --------------------- #
  g_H1_k = g_x * _I2 / {vdt}_sqrt  # shape (d, ..)
  H2 = x + g_H1_k  # shape (d, ..)
  H3 = x - g_H1_k  # shape (d, ..)

  g1 = g_x * _I1  # shape (d, ..)
  x_new = x + f_x + g1  # shape (d, ..)
  x_new += 0.5 * {vdt}_sqrt * g(H2, {", ".join(parameters)})
  x_new -= 0.5 * {vdt}_sqrt * g(H3, {", ".join(parameters)})
  '''
  code_splits = code_string.split('\n')
  code_lines.extend(code_splits)


def _srk1_wrapper(f, g, dt, sde_type, var_type, wiener_type, show_code, num_iter):
  vdt, variables, parameters, arguments, func_name = basic_info(f=f, g=g)

  # 1. code scope
  code_scope = {'f': f, 'g': g, vdt: dt, f'{vdt}_sqrt': dt ** 0.5,
                'math': math, 'num_iter': num_iter}

  # 2. code lines
  code_lines = [f'def {func_name}({", ".join(arguments)}):']

  if var_type == constants.SYSTEM_VAR:
    if len(variables) > 1:
      raise ValueError(f'SDE_INT with {constants.SYSTEM_VAR} variable type only '
                       f'supports one system variable. But we got {variables}.')

    if wiener_type == constants.SCALAR_WIENER:
      _srk1_system_var_with_scalar_wiener(sde_type, code_lines, variables, parameters, vdt)
    elif wiener_type == constants.VECTOR_WIENER:
      _srk1_system_var_with_vector_wiener(sde_type, code_lines, variables, parameters, vdt)
    else:
      raise ValueError(f'Unknown Wiener type: {wiener_type}, we only '
                       f'supports {constants.SUPPORTED_WIENER_TYPE}')

  elif var_type == constants.SCALAR_VAR:
    if wiener_type == constants.SCALAR_WIENER:
      _srk2_pop_or_scalar_var_scalar_wiener(sde_type, code_lines, variables, parameters, vdt)
    elif wiener_type == constants.VECTOR_WIENER:
      _srk1_scalar_var_with_vector_wiener(sde_type, code_lines, variables, parameters, vdt)
    else:
      raise ValueError(f'Unknown Wiener type: {wiener_type}, we only '
                       f'supports {constants.SUPPORTED_WIENER_TYPE}')

  elif var_type == constants.POP_VAR:
    if wiener_type == constants.SCALAR_WIENER:
      _srk2_pop_or_scalar_var_scalar_wiener(sde_type, code_lines, variables, parameters, vdt)
    elif wiener_type == constants.VECTOR_WIENER:
      _srk2_pop_var_vector_wiener(sde_type, code_lines, variables, parameters, vdt)
    else:
      raise ValueError(f'Unknown Wiener type: {wiener_type}, we only '
                       f'supports {constants.SUPPORTED_WIENER_TYPE}')

  else:
    raise ValueError(f'Unknown var type: {var_type}, we only '
                     f'supports {constants.SUPPORTED_VAR_TYPE}')
  # returns
  new_vars = [f'{var}_new' for var in variables]
  code_lines.append(f'  return {", ".join(new_vars)}')

  # return and compile
  utils.compile_code(code_lines, code_scope, show_code, variables)
  return code_scope[func_name]


def _srk2_wrapper():
  pass


def _wrap(wrapper, f, g, dt, sde_type, var_type, wiener_type, show_code, num_iter):
  """The base function to format a SRK method.

  Parameters
  ----------
  f : callable
      The drift function of the SDE_INT.
  g : callable
      The diffusion function of the SDE_INT.
  dt : float
      The numerical precision.
  sde_type : str
      "utils.ITO_SDE" : Ito's Stochastic Calculus.
      "utils.STRA_SDE" : Stratonovich's Stochastic Calculus.
  wiener_type : str
  var_type : str
      "scalar" : with the shape of ().
      "population" : with the shape of (N,) or (N1, N2) or (N1, N2, ...).
      "system": with the shape of (d, ), (d, N), or (d, N1, N2).
  show_code : bool
      Whether show the formatted code.

  Returns
  -------
  numerical_func : callable
      The numerical function.
  """

  sde_type = constants.ITO_SDE if sde_type is None else sde_type
  assert sde_type in constants.SUPPORTED_INTG_TYPE, f'Currently, BrainPy only support SDE_INT types: ' \
                                                    f'{constants.SUPPORTED_INTG_TYPE}. But we got {sde_type}.'

  var_type = constants.POP_VAR if var_type is None else var_type
  assert var_type in constants.SUPPORTED_VAR_TYPE, f'Currently, BrainPy only supports variable types: ' \
                                                   f'{constants.SUPPORTED_VAR_TYPE}. But we got {var_type}.'

  wiener_type = constants.SCALAR_WIENER if wiener_type is None else wiener_type
  assert wiener_type in constants.SUPPORTED_WIENER_TYPE, f'Currently, BrainPy only supports Wiener ' \
                                                         f'Process types: {constants.SUPPORTED_WIENER_TYPE}. ' \
                                                         f'But we got {wiener_type}.'

  show_code = False if show_code is None else show_code
  dt = math.get_dt() if dt is None else dt
  num_iter = 10 if num_iter is None else num_iter

  if f is not None and g is not None:
    return wrapper(f=f, g=g, dt=dt, show_code=show_code, sde_type=sde_type,
                   var_type=var_type, wiener_type=wiener_type, num_iter=num_iter)

  elif f is not None:
    return lambda g: wrapper(f=f, g=g, dt=dt, show_code=show_code, sde_type=sde_type,
                             var_type=var_type, wiener_type=wiener_type, num_iter=num_iter)

  elif g is not None:
    return lambda f: wrapper(f=f, g=g, dt=dt, show_code=show_code, sde_type=sde_type,
                             var_type=var_type, wiener_type=wiener_type, num_iter=num_iter)

  else:
    raise ValueError('Must provide "f" or "g".')


# ------------------
# Numerical methods
# ------------------


def srk1_strong(f=None, g=None, dt=None, sde_type=None, var_type=None, wiener_type=None, num_iter=None, show_code=None):
  return _wrap(_srk1_wrapper, f=f, g=g, dt=dt, sde_type=sde_type, var_type=var_type,
               wiener_type=wiener_type, show_code=show_code, num_iter=num_iter)


def srk2_strong(f=None, g=None, dt=None, sde_type=None, var_type=None, wiener_type=None, num_iter=None, show_code=None):
  return _wrap(_srk2_wrapper, f=f, g=g, dt=dt, sde_type=sde_type, var_type=var_type,
               wiener_type=wiener_type, show_code=show_code, num_iter=num_iter)



# -*- coding: utf-8 -*-

from brainpy.base.function import Function

__all__ = [
  'function',
]


def function(f=None, nodes=None, dyn_vars=None, name=None):
  if f is None:
    return lambda func: Function(f=func, nodes=nodes, dyn_vars=dyn_vars, name=name)

  else:
    if nodes is None:
      raise ValueError(f'"nodes" cannot be None when "f" is provided.')
    return Function(f=f, nodes=nodes, dyn_vars=dyn_vars, name=name)
